#!/usr/bin/env python
"""
barycenter_broaden_atmmodel
---------------------------
Doppler shift the Tapas atmosphere model and save to files.
This makes the RV precision calculations faster.

"""

import argparse
import sys
from os.path import join
from typing import List, Optional

import numpy as np

from eniric import config
from eniric.atmosphere import Atmosphere

choices = [None, "ALL"]
choices.extend(config.bands["all"])


def _parser():
    """Take care of all the argparse stuff."""
    parser = argparse.ArgumentParser(
        description=("Pre-doppler shift atmosphere masks.")
    )

    parser.add_argument(
        "-b",
        "--bands",
        type=str,
        default="ALL",
        choices=choices,
        help="Wavelength bands to select. Default is None.",
        nargs="+",
    )
    parser.add_argument("-v", "--verbose", help="Turn on verbose.", action="store_true")
    return parser.parse_args()


def main(bands: Optional[List[str]] = None, verbose: bool = False) -> None:
    """Preform the barycentric shifting of atmosphere masks and saves result.

    This saves time in the precision determination code.

    Parameters
    ----------
    bands: list of str
        Wavelength bands to perform barycenter shifts on. Default is all bands.

    """
    if (bands is None) or ("ALL" in bands):
        bands_ = config.bands["all"]
    else:
        bands_ = bands

    for band in bands_:
        unshifted_atmmodel = join(
            config.pathdir,
            config.paths["atmmodel"],
            "{0}_{1}.dat".format(config.atmmodel["base"], band),
        )
        if verbose:
            print("Reading atmospheric model...", unshifted_atmmodel)

        atm = Atmosphere.from_file(unshifted_atmmodel)

        if verbose:
            print("Calculating impact of Barycentric movement on mask...")
        org_mask = atm.mask
        masked_before = np.sum(org_mask)
        atm.barycenter_broaden(consecutive_test=True)

        masked_after = np.sum(atm.mask)
        if verbose:
            print(
                "Masked fraction before = {0:0.03f}".format(
                    (len(org_mask) - masked_before) / len(org_mask)
                )
            )
            print(
                "Masked fraction after = {0:0.03f}".format(
                    (len(atm.mask) - masked_after) / len(atm.mask)
                )
            )

        shifted_atmmodel = unshifted_atmmodel.replace(".dat", "_bary.dat")
        if verbose:
            print(
                "Saving doppler-shifted atmosphere model to {}".format(shifted_atmmodel)
            )

        header = ["# atm_wav(nm)", "atm_flux", "atm_std_flux", "atm_mask"]
        atm.to_file(fname=shifted_atmmodel, header=header, fmt="%11.8f")

    if verbose:
        print("Finished barycentric shifting of atmosphere masks")


if __name__ == "__main__":
    args = vars(_parser())
    opts = {k: args[k] for k in args}
    try:
        main(**opts)
        sys.exit(0)
    except:
        raise



# Inspired by Starfish

import os
import shutil

import oyaml as yaml

from . import DEFAULT_CONFIG_FILE


class Config(object):
    def __init__(self, path):
        """Creates a persistent Config object from the give file.

        This class is not meant to be instantiated by the User,
        but rather to be used via ``eniric.config``.

        Parameters
        ----------
        path: str or path-like
            The filename for creating the Config. Must be YAML.

        """
        self._path = path
        self._protect_rewrites = True

        with open(self._path, "r") as fd:
            self._config = yaml.safe_load(fd)

        self._protect_rewrites = os.path.abspath(path) == DEFAULT_CONFIG_FILE

    def get_pathdir(self):
        """Directory of the config file."""
        return os.path.dirname(self._path)

    pathdir = property(get_pathdir)

    def __contains__(self, key):
        return key in self._config

    def __delitem__(self, key):
        del self._config[key]

    def __eq__(self, other):
        return self._config.__eq__(other)

    def __getitem__(self, key):
        return self._config[key]

    def __setitem__(self, key, value):
        ret = self._config.__setitem__(key, value)
        return ret

    def __getattr__(self, key):
        if key in self:
            if key == "paths":
                paths = self["paths"]
                for k, value in paths.items():
                    if isinstance(value, list):
                        paths[k] = os.path.join(*value)
                return paths
            elif key == "cache":
                cache = self["cache"]
                if (cache["location"] is None) or (cache["location"] == "None"):
                    cache["location"] = None  # Disables caching
                elif isinstance(cache["location"], list):
                    cache["location"] = os.path.join(*cache["location"])
                return cache
            return self[key]
        else:
            return super().__getattribute__(key)

    def __setattr__(self, key, value):
        if key not in ["_path", "_protect_rewrites", "_config", "pathdir"]:
            if key in self:
                self.__setitem__(key, value)
                self._rewrite()

        super().__setattr__(key, value)

    def _rewrite(self):
        if self._protect_rewrites:
            raise RuntimeError(
                "The default file is not allowed to be overwritten. Please copy a file using "
                "config.copy_file(<path>) for your use."
            )
        with open(self._path, "w") as fd:
            yaml.safe_dump(self._config, fd)

    def update(self, d=None, **kwargs):
        """Update the config values and save to the config file."""
        if d is None:
            d = {}
        protected_rewrites = self._protect_rewrites
        self._protect_rewrites = True

        self._config.update(d, **kwargs)

        self._protect_rewrites = protected_rewrites

        self._rewrite()

    def change_file(self, filename):
        """Change the current configuration to use the given YAML file.

        Parameters
        ----------
        filename: str or path-like
            The YAML file to switch to using for config.

        Example
        -------
        .. code-block:: python

            eniric.config.change_file('new_config.yaml')

        """
        self._path = filename
        with open(self._path, "r") as fd:
            self._config = yaml.safe_load(fd)

    def copy_file(self, directory=None, switch=True):
        """Copies the master config file to the given directory.

        Parameters
        ----------
        directory: str or path-like
            The directory to copy the ``config.yaml`` file to. Default is os.getcwd().
        switch: bool
            If True, will switch the current config to use the copied file. Default is True

        Example
        -------
        .. code-block:: python

            eniric.config.copy_file()

        """
        if directory is None:
            directory = os.getcwd()
        outname = os.path.join(directory, "config.yaml")
        shutil.copy(DEFAULT_CONFIG_FILE, outname)
        if switch:
            self.change_file(outname)



"""Test eniric.precision."""

import astropy.units as u
import numpy as np
import pytest
from astropy import constants as const
from astropy.units import Quantity

from eniric.atmosphere import Atmosphere
from eniric.precision import (
    incremental_quality,
    incremental_rv,
    mask_check,
    pixel_weights,
    quality,
    rv_precision,
    sqrt_sum_wis,
)
from eniric.snr_normalization import snr_constant_band
from eniric.utilities import (
    band_limits,
    load_aces_spectrum,
    wav_selector,
    weighted_error,
)

m_per_s = u.meter / u.second
per_s_cm2 = (1 / u.second) / (u.centimeter ** 2)
c = const.c

xfail = pytest.mark.xfail


def test_rvprev_calc(test_spec, wav_unit, flux_unit, trans_unit):
    """Test that rv_precision can handle inputs as Quantities or unitless and returns a scalar Quantity."""
    wav = test_spec[0] * wav_unit
    flux = test_spec[1] * flux_unit
    mask = test_spec[2]
    if test_spec[2] is not None:
        mask *= trans_unit

    rv = rv_precision(wav, flux, mask)
    assert rv.unit == m_per_s
    assert not hasattr(rv.value, "__len__")  # assert value is a scalar
    assert isinstance(rv, u.Quantity)


def test_rvprev_calc_with_lists(test_spec):
    """Test that it can handle list input also."""
    wav = list(test_spec[0])
    flux = list(test_spec[1])
    mask = test_spec[2]
    rv = rv_precision(wav, flux, mask)
    assert not hasattr(rv.value, "__len__")  # assert value is a scalar
    assert isinstance(rv, u.Quantity)
    assert rv.unit == m_per_s


def test_sqrt_sum_wis_with_no_units(test_spec):
    """Test that sqrt_sum_wis can handle inputs as Quantities or unitless.
    Returns a dimensionless unscaled Quantity.
    """
    sqrtsumwis = sqrt_sum_wis(test_spec[0], test_spec[1], test_spec[2])
    # Doesn't turn into quantity if does not have to.
    assert not isinstance(sqrtsumwis, u.Quantity)
    assert not hasattr(sqrtsumwis, "__len__")  # assert value is a scalar


def test_sqrt_sum_wis(test_spec, wav_unit, flux_unit, trans_unit):
    """Test that sqrt_sum_wis can handle inputs as Quantities or unitless.

    Returns a dimensionless unscaled Quantity.
    """
    wav = test_spec[0] * wav_unit
    flux = test_spec[1] * flux_unit
    mask = test_spec[2]
    if test_spec[2] is not None:
        mask *= trans_unit

    sqrtsumwis = sqrt_sum_wis(wav, flux, mask)
    print("wav", wav, type(wav))
    print("flux", flux, type(flux))
    print("mask", mask, type(mask))
    print("sqrtsumwis", sqrtsumwis, type(sqrtsumwis))
    if (
        (isinstance(wav, Quantity))
        or (isinstance(flux, Quantity))
        or (isinstance(mask, Quantity) and (test_spec[2] is not None))
    ):
        assert isinstance(sqrtsumwis, u.Quantity)
        # Check is unscaled and dimensionless Quantity
        assert sqrtsumwis.unit == u.dimensionless_unscaled
        assert not hasattr(sqrtsumwis.value, "__len__")  # assert value is a scalar
    else:
        assert not hasattr(sqrtsumwis, "__len__")  # assert value is a scalar


def test_relation_of_rv_to_sqrtsumwis(test_spec, wav_unit, flux_unit, trans_unit):
    """Test relation of sqrtsumwis to rv_precision."""
    wav = test_spec[0] * wav_unit
    flux = test_spec[1] * flux_unit
    mask = test_spec[2]
    if test_spec[2] is not None:
        mask *= trans_unit
        mask = mask ** 2
    assert np.all(
        rv_precision(wav, flux, mask=mask) == c / sqrt_sum_wis(wav, flux, mask=mask)
    )


def test_transmission_reduces_precision(test_spec):
    """Check that a transmission vector reduces precision calculation."""
    wav = test_spec[0]
    flux = test_spec[1]
    transmission = test_spec[2]

    # Value should be less then normal if trans <=1
    if transmission is not None:
        assert rv_precision(wav, flux, mask=None) < rv_precision(
            wav, flux, mask=transmission
        )
    # mask=None is the same as mask of all 1.
    assert rv_precision(wav, flux, mask=None) == rv_precision(
        wav, flux, mask=np.ones_like(wav)
    )


def test_both_gradients_possible(test_spec):
    wav = test_spec[0]
    flux = test_spec[1]
    transmission = test_spec[2]
    __ = rv_precision(wav, flux, mask=transmission, grad=False).value
    __ = rv_precision(wav, flux, mask=transmission, grad=True).value
    assert True


@pytest.mark.parametrize("scale", [0.1, 1, 2, 100, 0.1, 0.5])
def test_quality_independent_of_flux_level(scale):
    """Q of a spectrum is independent of flux level."""
    wavelength = np.arange(100)
    flux = np.random.random(100)
    assert np.allclose(quality(wavelength, flux), quality(wavelength, flux * scale))


def test_quality_independent_of_units(test_spec, wav_unit, flux_unit):
    """Quality should be returned as unitless (not a quantity)."""
    wave = test_spec[0] * wav_unit
    flux = test_spec[1] * flux_unit
    q = quality(wave, flux)

    assert not isinstance(q, Quantity)
    assert isinstance(q, float)
    assert not hasattr(q, "__len__")  # assert value is a scalar


@pytest.mark.parametrize(
    "wav_unit2, flux_unit2, quantity",
    ([(u.nanometer, 1 / u.second, True), (u.meter, u.erg, True), (1, 1, False)]),
)
@pytest.mark.parametrize(
    "wave, flux, expected",
    [
        ([1, 2, 3], [1, 2, 3], [1, 2, 3]),
        (
            [2.2, 2.3, 2.4],
            [0.99, 0.97, 0.7],
            [0.195_555_556, 11.466_211_34, 59.986_857_1],
        ),
    ],
)
def test_pixel_weights_gradient_with_fixed_values(
    wave, flux, expected, wav_unit2, flux_unit2, quantity
):
    """Pixel_weights with gradient formulation.
    Change detector with some simple numbers.
    """
    result = pixel_weights(wave * wav_unit2, flux * flux_unit2, grad=True)
    assert np.allclose(result, expected)
    # Checks quantities also work
    if quantity:
        assert isinstance(result, Quantity)
        assert result.unit == u.dimensionless_unscaled


@pytest.mark.parametrize(
    "wav_unit2, flux_unit2, quantity",
    [(u.nanometer, 1 / u.second, True), (u.meter, u.erg, True), (1, 1, False)],
)
@pytest.mark.parametrize(
    "wave, flux, expected",
    [
        ([1, 2, 3], [1, 2, 3], [1, 2]),
        ([2.2, 2.3, 2.4], [0.99, 0.97, 0.7], [0.195_555_556, 39.756_804_12]),
    ],
)
def test_pixel_weights(wave, flux, expected, wav_unit2, flux_unit2, quantity):
    """Pixel_weights with finite difference formulation.
    Change detector with some simple numbers.
    """
    print(wave, flux, expected, wav_unit2, flux_unit2, quantity)
    result = pixel_weights(wave * wav_unit2, flux * flux_unit2, grad=False)
    assert np.allclose(result, expected)
    # Checks quantities also work
    if quantity:
        assert isinstance(result, Quantity)
        assert result.unit == u.dimensionless_unscaled


@pytest.mark.parametrize(
    "mask",
    [np.array([1, 0, -0.1]), np.array([1.1, 0, 1])],  # lower than 0  # Greater than 1
)
def test_mask_check(mask):
    with pytest.raises(ValueError):
        mask_check(mask)


@pytest.mark.parametrize(
    "mask", [np.array([1, 0, 1, 0, 0.2, 1, 0.4]) * u.m, np.array([1, 0, 1]) * u.s]
)
def test_mask_check_type_error(mask):
    with pytest.raises(TypeError):
        mask_check(mask)


@pytest.mark.parametrize("trans_unit2", [m_per_s, per_s_cm2, u.meter])
def test_sqrt_sum_wis_with_mask_with_unit_fails(
    test_spec, wav_unit, flux_unit, trans_unit2
):
    """Assert a transmission with a unit fails with type error."""
    wav = test_spec[0] * wav_unit
    flux = test_spec[1] * flux_unit
    transmission = np.random.rand(len(wav)) * trans_unit2

    with pytest.raises(TypeError):
        sqrt_sum_wis(wav, flux, mask=transmission ** 2)

    with pytest.raises(TypeError):
        rv_precision(wav, flux, mask=transmission ** 2)


def test_sqrt_sum_wis_transmission_outofbounds(test_spec, wav_unit, flux_unit):
    """Transmission must be within 0-1."""
    wav = test_spec[0] * wav_unit
    flux = test_spec[1] * flux_unit
    mask_1 = np.random.randn(len(wav))
    mask_2 = np.random.rand(len(wav))

    mask_1[0] = 5  # Outside 0-1
    mask_2[-1] = -2  # Outside 0-1

    # Higher value
    with pytest.raises(ValueError):
        rv_precision(wav, flux, mask=mask_1)

    with pytest.raises(ValueError):
        sqrt_sum_wis(wav, flux, mask=mask_1)

        # Lower value
    with pytest.raises(ValueError):
        sqrt_sum_wis(wav, flux, mask=mask_2)

    with pytest.raises(ValueError):
        sqrt_sum_wis(wav, flux, mask=mask_2)


def test_sqrtsumwis_warns_nonfinite(grad_flag):
    """Some warning tests."""
    with pytest.warns(UserWarning, match="This will cause infinite errors."):
        sqrt_sum_wis(
            np.array([1, 2, 3]),
            np.array([1, 2, 3]),
            np.array([0, 0, 0]),
            grad=grad_flag,
        )  # All masked

    with pytest.warns(UserWarning, match="Weight sum is not finite"):
        sqrt_sum_wis(
            np.array([2, 2, 2]),
            np.array([1, 2, 3]),
            np.array([1, 1, 1]),
            grad=grad_flag,
        )  # infinite gradient


@pytest.fixture(params=[1, 2, 5])
def increment_percent(request):
    return request.param


@pytest.fixture(params=["K", "J"])
def real_spec(request, use_test_config):
    band = request.param
    wav, flux = load_aces_spectrum([3900, 4.5, 0.0, 0])
    wav, flux = wav_selector(wav, flux, *band_limits(band))
    flux = flux / snr_constant_band(wav, flux, 100, band)
    atm = Atmosphere.from_band(band).at(wav)
    return wav, flux, atm.transmission


@xfail(raises=ModuleNotFoundError, reason="Issue with Starfish install.")
def test_increment_quality_gives_reasonable_length(real_spec, increment_percent):
    """The expected number of steps would be between the
    wavelength difference divided by the
    first and last point * the percent increment.
    """
    print(real_spec)
    wav, flux = real_spec[0], real_spec[1]
    x, q = incremental_quality(wav, flux, percent=increment_percent)
    d1 = wav[0] * increment_percent / 100
    d2 = wav[-1] * increment_percent / 100
    dlambda = wav[-1] - wav[0]
    len_q = len(q)

    assert len_q >= np.floor(dlambda / d1)
    assert len_q <= np.ceil(dlambda / d2 + 1)
    assert len(x) == len_q


@xfail(raises=ModuleNotFoundError, reason="Issue with Starfish install.")
def test_increments_rv_gives_reasonable_length(real_spec, increment_percent):
    """The expected number of steps would be between the
     wavelength difference divided by the
     first and last point * the percent increment.
     """
    wav, flux, mask = real_spec[0], real_spec[1], real_spec[2]
    x, rv = incremental_rv(wav, flux, mask=mask, percent=increment_percent)
    d1 = wav[0] * increment_percent / 100
    d2 = wav[-1] * increment_percent / 100
    dlambda = wav[-1] - wav[0]
    len_rv = len(rv)

    assert len_rv >= np.floor(dlambda / d1)
    assert len_rv <= np.ceil(dlambda / d2 + 1)
    assert len(x) == len_rv


@xfail(raises=ModuleNotFoundError, reason="Issue with Starfish install.")
@pytest.mark.parametrize("no_mask", [True, False])
def test_increments_rv_accumulate_same_as_full(real_spec, increment_percent, no_mask):
    """Assuming that the weighted rv from the steps should equal the rv from the band."""
    wav, flux, mask = real_spec[0], real_spec[1], real_spec[2]
    if no_mask:
        # Try with mask= None also.
        mask = None

    rv_full = rv_precision(wav, flux, mask=mask).value
    x, incremented_rv = incremental_rv(wav, flux, mask=mask, percent=increment_percent)
    incremented_weighted = weighted_error(incremented_rv)

    assert np.round(rv_full, 2) == np.round(incremented_weighted, 2)
    assert x[0] > wav[0]
    assert [-1] < wav[-1]



"""Spectral broadening functions:

Used to convolve the model spectra for
    - Stellar rotation broadening.
    - Instrumental profile broadening.

Uses joblib.Memory to cache convolution results to skip repeated computation.
"""
from typing import Optional, Union

import joblib
import numpy as np
from astropy import constants as const
from joblib import Memory, Parallel, delayed
from numpy.core.multiarray import ndarray
from tqdm import tqdm

from eniric import config
from eniric.utilities import band_selector, cpu_minus_one, mask_between, wav_selector

# Cache convolution results.
memory = Memory(location=config.cache["location"], verbose=0)

c_kmps = const.c.to("km/s").value

num_cpu_minus_1 = cpu_minus_one()


@memory.cache(ignore=["num_procs", "verbose"])
def rotational_convolution(
    wavelength: ndarray,
    extended_wav: ndarray,
    extended_flux: ndarray,
    vsini: float,
    *,
    epsilon: float = 0.6,
    normalize: bool = True,
    num_procs: Optional[Union[int, joblib.parallel.Parallel]] = None,
    verbose: bool = True,
) -> ndarray:
    r"""Perform Rotational convolution.

    Parameters
    ----------
    wavelength: ndarray
        Wavelength array.
    extended_wav: ndarray
       Extended wavelength array to avoid boundary issues.
    extended_flux: ndarray
        Photon flux for the extended wavelength array.
    vsini: float
        Rotational velocity in km/s.
    epsilon: float
        Limb darkening coefficient. Default is 0.6.
    normalize: bool
        Area normalize the broadening kernel. This corrects for kernel area with unequal wavelength spacing.
        Default is True.
    num_procs: int, None or joblib.parallel.Parallel.
        Number of processes to use, n_job parameter in joblib.
        If num_procs =  1, then a single core is used.
        Can also be a joblib.parallel.Parallel instance. Default is None.
    verbose: bool
        Show the tqdm progress bar. Default is True.

    Returns
    -------
    convolved_flux: ndarray
        The convolved flux evaluated at the wavelength array.

    """

    def element_rot_convolution(single_wav: float) -> float:
        r"""Embarrassingly parallel part of rotational convolution.

        Calculates the convolution value for a single pixel.
        The parameters extended_wav, extended_flux, vsini, epsilon and normalize
        are obtained from the outer scope.

        Parameters
        ----------
        single_wav: float
            Wavelength to calculate the convolution at.

        Returns
        -------
        sum_val: float
            Sum of flux convolved for this wavelength.

        """
        # Select all values such that they are within the fwhm limits
        delta_lambda_l = single_wav * vsini / c_kmps

        index_mask = mask_between(
            extended_wav, single_wav - delta_lambda_l, single_wav + delta_lambda_l
        )

        flux_2convolve = extended_flux[index_mask]
        rotation_profile = rotation_kernel(
            extended_wav[index_mask] - single_wav,
            delta_lambda_l,
            vsini=vsini,
            epsilon=epsilon,
        )

        sum_val = np.sum(rotation_profile * flux_2convolve)

        if normalize:
            # Correct for the effect of non-equidistant sampling
            unitary_rot_val = np.sum(rotation_profile)
            sum_val /= unitary_rot_val

        return sum_val

    if num_procs is None:
        num_procs = num_cpu_minus_1

    if vsini != 0:
        tqdm_wav = tqdm(wavelength, disable=not verbose)

        if isinstance(num_procs, int):
            with Parallel(n_jobs=num_procs) as parallel:
                convolved_flux = np.asarray(
                    parallel(delayed(element_rot_convolution)(wav) for wav in tqdm_wav)
                )
        else:
            try:
                # Assume num_procs is a joblib.parallel.Parallel.
                convolved_flux = np.asarray(
                    num_procs(delayed(element_rot_convolution)(wav) for wav in tqdm_wav)
                )
            except TypeError:
                raise TypeError(
                    "num_proc must be an int or joblib.parallel.Parallel. Not '{}'".format(
                        type(num_procs)
                    )
                )
    else:
        # Skip convolution for vsini=0
        if wavelength is extended_wav:
            convolved_flux = extended_flux  # No change
        else:
            # Interpolate to the new wavelength vector.
            convolved_flux = np.interp(wavelength, extended_wav, extended_flux)
    return convolved_flux


@memory.cache(ignore=["num_procs", "verbose"])
def resolution_convolution(
    wavelength: ndarray,
    extended_wav: ndarray,
    extended_flux: ndarray,
    R: float,
    *,
    fwhm_lim: float = 5.0,
    normalize: bool = True,
    num_procs: Optional[Union[int, joblib.parallel.Parallel]] = None,
    verbose: bool = True,
) -> ndarray:
    r"""Perform Resolution convolution.

    Parameters
    ----------
    wavelength: ndarray
        Wavelength array.
    extended_wav: ndarray
       Extended wavelength array to avoid boundary issues.
    extended_flux: ndarray
        Photon flux for the extended wavelength array.
    R: float
        Resolution of Guassian instrumental profile.
    fwhm_lim: float
        FWHM limit for instrument broadening. Default is 5.0.
    normalize: bool
        Area normalize the broadening kernel. This corrects for kernel area with unequal wavelength spacing.
        Default is True.
    num_procs: int, None or joblib.parallel.Parallel.
        Number of processes to use, n_job parameter in joblib.
        If num_procs =  1, then a single core is used.
        Can also be a joblib.parallel.Parallel instance.
    verbose: bool
        Show the tqdm progress bar. Default is True.

    Returns
    -------
    convolved_flux: ndarray
        The convolved flux evaluated at the wavelength array.
    """

    def element_res_convolution(single_wav: float) -> float:
        r"""Embarrassingly parallel component of resolution convolution.

        Calculates the convolution value for a single pixel.

        The parameters extended_wav, fwhm_lim, R and normalize
        are obtained from the outer scope.

        Parameters
        ----------
        single_wav: float
            Wavelength value to calculate convolution at.

        Returns
        -------
        sum_val: float
            Sum of flux convolved for this pixel/wavelength.
        """
        fwhm = single_wav / R
        # Mask of wavelength range within fwhm_lim* fwhm of wav
        fwhm_space = fwhm_lim * fwhm
        index_mask = mask_between(
            extended_wav, single_wav - fwhm_space, single_wav + fwhm_space
        )

        flux_2convolve = extended_flux[index_mask]
        # Gaussian Instrument Profile for given resolution and wavelength
        instrument_profile = unitary_gaussian(
            extended_wav[index_mask], single_wav, fwhm=fwhm
        )

        sum_val = np.sum(instrument_profile * flux_2convolve)
        if normalize:
            # Correct for the effect of convolution with non-equidistant positions
            unitary_val = np.sum(instrument_profile)
            sum_val /= unitary_val

        return sum_val

    tqdm_wav = tqdm(wavelength, disable=not verbose)

    if num_procs is None:
        num_procs = num_cpu_minus_1

    if isinstance(num_procs, int):
        with Parallel(n_jobs=num_procs) as parallel:
            convolved_flux = np.asarray(
                parallel(delayed(element_res_convolution)(wav) for wav in tqdm_wav)
            )
    else:
        # Assume num_procs is joblib.parallel.Parallel.
        try:
            convolved_flux = np.asarray(
                num_procs(delayed(element_res_convolution)(wav) for wav in tqdm_wav)
            )
        except TypeError:
            raise TypeError(
                "num_proc must be an int or joblib.parallel.Parallel. Not '{}'".format(
                    type(num_procs)
                )
            )
    return convolved_flux


@memory.cache(ignore=["num_procs", "verbose"])
def convolution(
    wav: ndarray,
    flux: ndarray,
    vsini: float,
    R: float,
    band: str = "All",
    *,
    epsilon: float = 0.6,
    fwhm_lim: float = 5.0,
    num_procs: Optional[Union[int, joblib.parallel.Parallel]] = None,
    normalize: bool = True,
    verbose: bool = True,
):
    r"""Perform rotational then Instrumental broadening with convolutions.

    Parameters
    ----------
    wav: ndarray
        Wavelength array.
    flux: ndarray
        Flux array.
    vsini: float
        Rotational velocity in km/s.
    R: int
        Resolution of instrumental profile.
    band: str
        Wavelength band to choose, default is "All".
    epsilon: float
        Limb darkening coefficient. Default is 0.6.
    fwhm_lim: float
        FWHM limit for instrument broadening. Default is 5.0.
    normalize: bool
        Area normalize the broadening kernel. This corrects for kernel area with unequal wavelength spacing.
        Default is True.
    num_procs: int, None or joblib.parallel.Parallel.
        Number of processes to use, n_job parameter in joblib.
        If num_procs =  1, then a single core is used.
        Can also be a joblib.parallel.Parallel instance.
    verbose: bool
        Show the tqdm progress bar. Default is True.

    Returns
    -------
    wav_band: ndarray
        Wavelength for the selected band.
    flux_band: ndarray
        Original flux for the selected band.
    flux_conv: ndarray
        Convolved flux for the selected band.
    """

    wav_band, flux_band = band_selector(wav, flux, band)

    # Calculate FWHM at each end for the convolution
    fwhm_min = wav_band[0] / R  # fwhm at the extremes of vector
    fwhm_max = wav_band[-1] / R

    # performing convolution with rotation kernel
    if verbose:
        print("Starting the Rotation convolution for vsini={0:.2f}...".format(vsini))

    delta_lambda_min = wav_band[0] * vsini / c_kmps
    delta_lambda_max = wav_band[-1] * vsini / c_kmps

    # widest wavelength bin for the rotation convolution
    lower_lim = wav_band[0] - delta_lambda_min - fwhm_lim * fwhm_min
    upper_lim = wav_band[-1] + delta_lambda_max + fwhm_lim * fwhm_max
    wav_ext_rotation, flux_ext_rotation = wav_selector(wav, flux, lower_lim, upper_lim)

    # wide wavelength bin for the resolution_convolution
    lower_lim = wav_band[0] - fwhm_lim * fwhm_min
    upper_lim = wav_band[-1] + fwhm_lim * fwhm_max
    extended_wav, __ = wav_selector(wav, flux, lower_lim, upper_lim)

    # rotational convolution
    flux_conv_rot = rotational_convolution(
        extended_wav,
        wav_ext_rotation,
        flux_ext_rotation,
        vsini,
        epsilon=epsilon,
        num_procs=num_procs,
        normalize=normalize,
        verbose=verbose,
    )
    if verbose:
        print("Starting the Resolution convolution...")

    flux_conv_res = resolution_convolution(
        wav_band,
        extended_wav,
        flux_conv_rot,
        R,
        fwhm_lim=fwhm_lim,
        num_procs=num_procs,
        normalize=normalize,
        verbose=verbose,
    )

    return wav_band, flux_band, flux_conv_res


def unitary_gaussian(
    x: Union[range, int, ndarray],
    center: Union[float, int, str],
    fwhm: Union[float, int, str],
) -> ndarray:
    """Gaussian kernal of area 1.

    Parameters
    ----------
    x: array-like
        Position array.
    center: float
        Central position of Gaussian.
    fwhm: float
        Full Width at Half Maximum.

    Returns
    -------
    kernel: array-like
        Gaussian kernel.
    """
    if not isinstance(fwhm, (np.float, np.int)):
        raise TypeError("The fwhm value is not a number, {0}".format(type(fwhm)))
    if not isinstance(center, (np.float, np.int)):
        raise TypeError("The center value is not a number, {0}".format(type(center)))
    if not isinstance(x, np.ndarray):
        raise TypeError

    sigma = np.abs(fwhm) / (2 * np.sqrt(2 * np.log(2)))
    amp = 1.0 / (sigma * np.sqrt(2 * np.pi))
    tau = -((x - center) ** 2) / (2 * (sigma ** 2))
    kernel = amp * np.exp(tau)

    return kernel


def rotation_kernel(
    delta_lambdas: ndarray, delta_lambda_l: float, vsini: float, epsilon: float
) -> ndarray:
    r"""Rotation kernel for a given line center.

    Parameters
    ----------
    delta_lambdas: array
        Wavelength difference from the line center lambda.
    delta_lambda_l: float
        Maximum line shift of line center by vsini.
    vsini: float
        Projected rotational velocity in km/s.
    epsilon: float
        Linear limb-darkening coefficient, between 0 and 1.

    Returns
    -------
    kernel: array
        Rotational kernel.

    Notes
    -----
    Gray, D. F. (2005). The Observation and Analysis of Stellar Photospheres. 3rd ed. Cambridge University Press.

    """
    denominator = np.pi * vsini * (1.0 - epsilon / 3.0)
    lambda_ratio_sqr = (delta_lambdas / delta_lambda_l) ** 2.0

    c1 = 2.0 * (1.0 - epsilon) / denominator
    c2 = 0.5 * np.pi * epsilon / denominator
    kernel = c1 * np.sqrt(1.0 - lambda_ratio_sqr) + c2 * (1.0 - lambda_ratio_sqr)

    return kernel


def oned_circle_kernel(x: ndarray, center: float, fwhm: float):
    """Calculate the convolution kernel for a circular fiber.

    Parameters
    ----------
    x: array
        Value to evaluate kernel at.
    center: float
        Center of kernel.
    fwhm: float
        FWHM of desired kernel.

    Returns
    -------
    Collapsed circle kernel.

    Notes
    -----
    Tries to represent the broadening by the fiber of a fiber feed spectrograph.

    Artigau 2018 - stated this is mathematically equivalent to a cosine between -pi/2 and pi/2.
    """
    fwhm_scale = 2.094_395_1  # Numerically derived

    A = 1  # Amplitude
    B = fwhm_scale / fwhm  # Scale to give specific fwhm

    kernel = A * np.cos(B * (x - center))

    # Limit to main cosine lobe only
    upper_xi = center + np.pi / 2 / B
    lower_xi = center - np.pi / 2 / B
    mask = mask_between(x, lower_xi, upper_xi)
    kernel[~mask] = 0

    return kernel



#!/usr/bin/env python
"""
split_atmmodel
--------------
Script to split the large atmospheric model transmission spectra into the separate bands.
This create smaller files to load for each band for individual bands only.

"""
import argparse
import os
import sys
from os.path import join
from typing import List, Optional

import numpy as np

from eniric import config
from eniric.atmosphere import Atmosphere
from eniric.utilities import band_limits, doppler_shift_wav

atmmodel = "{0}.dat".format(config.atmmodel["base"])
choices = ["ALL"]
choices.extend(config.bands["all"])


def _parser():
    """Take care of all the argparse stuff."""
    parser = argparse.ArgumentParser(description="Band separate out atmospheric model.")

    parser.add_argument("-m", "--model", help="Model name", type=str, default=atmmodel)
    parser.add_argument(
        "-b",
        "--bands",
        type=str,
        default="ALL",
        nargs="+",
        choices=choices,
        help="Wavelength band to select, Default='All'",
    )
    parser.add_argument(
        "-d",
        "--data_dir",
        help="Telluric model data directory",
        type=str,
        default=join(config.pathdir, config.paths["atmmodel"]),
    )
    parser.add_argument(
        "--new_name",
        default=None,
        type=str,
        help="Base name for new files. Default is the original model name.",
    )
    parser.add_argument(
        "--rv_extend",
        default=100,
        type=check_positive,
        help="RV in km/s to extend the wavelength limits of the band. Default is 100.",
    )
    parser.add_argument(
        "-c",
        "--cutoff_depth",
        default=2,
        type=float,
        help=r"Telluric line depth cutoff percent. Default is 2.0.",
    )
    parser.add_argument("-v", "--verbose", help="Turn on verbose.", action="store_true")
    return parser.parse_args()


def check_positive(value: str) -> float:
    """Check if input is positive.

    http://stackoverflow.com/questions/14117415.

    Parameters
    ----------
    value: str
        A input string from argparse to check if it is a positive number.

    Returns
    -------
    float_value: float
        The value if it is positive as a float.

    Raises
    ------
    ArgumentTypeError:
        If value is not a positive number.
    """
    if not isinstance(value, str):
        raise ValueError("Input value is not a string.")

    float_value = float(value)
    if float_value <= 0:
        raise argparse.ArgumentTypeError(
            "{0:s} is an invalid positive value".format(value)
        )
    return float_value


def main(
    model: str = atmmodel,
    bands: Optional[List[str]] = None,
    new_name: Optional[str] = None,
    data_dir: Optional[str] = None,
    rv_extend: float = 100.0,
    cutoff_depth: float = 2.0,
    verbose: bool = False,
):
    """Split the large atmospheric model transmission spectra into the separate bands.

    Keeps wavelength of atmosphere model as nanometers.

    Parameters
    ----------
    model: str
        Telluric model file to load. It has columns wavelength, flux, std_flux, mask.
    bands: list[str]
        List bands to split model into separate files.
    new_name: str
        New file name base.
    data_dir: Optional[str]
        Directory for results. Can also be given in config.yaml "paths:atmmodel:"...
    rv_extend: float
        Absolute RV to extend wavelength range of telluric band.
        To later apply barycenter shifting. Default is 100.
    cutoff_depth: float
       Telluric line depth cutoff percent. Default = 2.0.

    Returns
    -------
    exit_status: int
        Unix-like exit. Non-zero indicates a failure occurred.

    """
    if (bands is None) or ("ALL" in bands):
        bands_ = config.bands["all"]
    else:
        bands_ = bands

    if new_name is None:
        new_name = model.split(".")[0]
    if data_dir is None:
        data_dir_ = join(config.pathdir, config.paths["atmmodel"])
    else:
        data_dir_ = str(data_dir)

    model_name = join(data_dir_, model)

    # If trying to obtain the provided model extract from and it doesn't yet exist
    # extract from tar.gz file. (Extracted it is 230 MB which is to large for Git)
    if "Average_TAPAS_2014.dat" == atmmodel:
        if not os.path.exists(model_name):
            if verbose:
                print("Unpacking Average_TAPAS_2014.dat.tar.gz...")
            import tarfile

            with tarfile.open(str(model_name) + ".tar.gz", "r") as tar:
                tar.extractall(data_dir_)
            if verbose:
                print("Unpacked")
    if verbose:
        print("Loading from_file {0}".format(model_name))
    atm = Atmosphere.from_file(model_name)

    # Return value from saving each band
    write_status = np.empty_like(bands_, dtype=int)

    for i, band in enumerate(bands_):
        if verbose:
            print("Starting {0}".format(band))
        filename_band = "{0}_{1}.dat".format(new_name, band)
        band_min, band_max = band_limits(band)

        band_min = doppler_shift_wav(band_min, -rv_extend)
        band_max = doppler_shift_wav(band_max, rv_extend)

        split_atm = atm.wave_select(band_min, band_max)

        # Apply telluric line mask
        atm.mask_transmission(depth=cutoff_depth)

        # Save the result to file
        filename = join(data_dir_, filename_band)
        header = ["# atm_wav(nm)", "atm_flux", "atm_std_flux", "atm_mask"]
        if verbose:
            print("Saving to_file {}".format(filename))
        write_status[i] = split_atm.to_file(filename, header=header, fmt="%11.8f")
    if verbose:
        print("Finished telluric model splitting")

    exit_status = int(
        np.sum(write_status)
    )  # If any extracts fail they will turn up here.
    return exit_status


if __name__ == "__main__":
    args = vars(_parser())
    opts = {k: args[k] for k in args}
    sys.exit(main(**opts))



import numpy as np
import pytest

import eniric.snr_normalization as snrnorm
import eniric.utilities as utils
from scripts.phoenix_precision import convolve_and_resample

xfail = pytest.mark.xfail


@xfail(raises=ModuleNotFoundError, reason="Issue with Starfish install.")
@pytest.mark.parametrize("desired_snr", [100.0, 150.0])
@pytest.mark.parametrize("band", ["J", "Y", "VIS"])
def test_snr_normalization(desired_snr, band, testing_spectrum):
    """Test SNR after normalizing function is the desired value.

    Testing on middle of J band.
    """
    wav, flux = testing_spectrum
    band_mid = utils.band_middle(band)

    # Searching for the closest index to 1.25
    index_reference = np.searchsorted(wav, [band_mid])[0]
    snr_estimate = np.sqrt(np.sum(flux[index_reference - 1 : index_reference + 2]))

    assert round(snr_estimate, 1) != desired_snr  # Assert SNR is not correct

    norm_const = snrnorm.snr_constant_band(wav, flux, snr=desired_snr, band=band)

    new_flux = flux / norm_const
    new_snr_estimate = np.sqrt(
        np.sum(new_flux[index_reference - 1 : index_reference + 2])
    )

    assert round(new_snr_estimate, 0) == desired_snr


@xfail(raises=ModuleNotFoundError, reason="Issue with Starfish install.")
@pytest.mark.parametrize("desired_snr", [50.0, 200.0])
@pytest.mark.parametrize("band", ["H", "GAP", "K"])
def test_snr_normalization_constant(desired_snr, band, testing_spectrum):
    """Test snr_constant_band and snr_constant_wav produce same result."""
    wav, flux = testing_spectrum
    band_mid = utils.band_middle(band)

    assert snrnorm.snr_constant_band(
        wav, flux, band=band, snr=desired_snr
    ) == snrnorm.snr_constant_wav(wav, flux, band_mid, snr=desired_snr)


@xfail(raises=ModuleNotFoundError, reason="Issue with Starfish install.")
@pytest.mark.parametrize("sampling", [3, 4.1])
def test_band_snr_norm(testing_spectrum, sampling):
    """Compared to wav snr norm."""
    wav, flux = testing_spectrum
    wav, flux = convolve_and_resample(
        wav, flux, vsini=1, R=100_000, band="J", sampling=sampling
    )
    assert snrnorm.snr_constant_band(
        wav, flux, band="J", snr=100, sampling=sampling
    ) == snrnorm.snr_constant_wav(wav, flux, wav_ref=1.25, snr=100, sampling=sampling)

    assert snrnorm.snr_constant_band(
        wav, flux, band="J", snr=100, sampling=sampling, verbose=True
    ) != snrnorm.snr_constant_wav(
        wav, flux, wav_ref=1.24, snr=100, sampling=sampling, verbose=True
    )


def test_sampling_index():
    """Some hard coded examples of sampling index."""
    # odd number
    assert snrnorm.sampling_index(6, 1) == [6]
    assert np.all(snrnorm.sampling_index(100, 3) == [99, 100, 101])
    assert np.all(snrnorm.sampling_index(5, 5) == [3, 4, 5, 6, 7])

    # even number
    assert np.all(snrnorm.sampling_index(10, 4) == [8, 9, 10, 11])
    assert np.all(snrnorm.sampling_index(10, 2) == [9, 10])
    # number that is at end of array.


@pytest.mark.parametrize("sample", [1, 2, 3, 4, 5, 7, 10, 15])
def test_sampling_size_and_type(sample):
    # Check number of values correct.
    for sample in [1, 2, 3, 4, 5, 7, 10, 15]:
        assert len(snrnorm.sampling_index(20, sample)) == sample
        assert isinstance(
            snrnorm.sampling_index(20, sample)[0], (int, np.integer)
        )  # index values must be int


def test_sampling_index_array():
    """Sampling index when array_length is given, or when index goes out of bounds."""
    assert np.all(snrnorm.sampling_index(100, 3, array_length=200) == [99, 100, 101])

    with pytest.raises(ValueError):
        snrnorm.sampling_index(3, 10)  # index will be < 0
    with pytest.raises(ValueError):
        snrnorm.sampling_index(3, 9, array_length=50)  # index will be < 0
    with pytest.raises(ValueError):
        snrnorm.sampling_index(
            46, 10, array_length=50
        )  # an index will be > (array_length - 1)


@pytest.mark.parametrize("band", ["VIS", "Z", "NIR", "J", "Y"])
def test_snr_constant_band_returns_mid_value_const(band):
    size = 100
    np.random.seed(40)
    flux = 500 * np.random.rand(
        size
    )  # To give a random spectrum (but consistent between tests)
    lim = utils.band_limits(band)
    wav = np.linspace(lim[0], lim[1], size)

    band_const = snrnorm.snr_constant_band(wav, flux, band=band)
    wav_const = snrnorm.snr_constant_wav(wav, flux, wav_ref=utils.band_middle(band))

    assert isinstance(band_const, float)
    assert isinstance(wav_const, float)
    assert band_const == wav_const  # Since band calls wave at midpoint


@pytest.mark.parametrize("band", ["VIS", "K", "H"])
@pytest.mark.parametrize("verbose", [True, False])
def test_snr_normalization_logic(band, verbose):
    """Testing direct value.

    snr = sqrt(sum(3 pixels))
    const = (snr/ref_snr)**2
    if pixel value = 3 then  the normalization constant will be 1.
    """
    size = 100
    band = "K"
    lim = utils.band_limits(band)
    wav = np.linspace(lim[0], lim[1], size)
    flux = 3 * np.ones(size)
    band_const = snrnorm.snr_constant_band(wav, flux, snr=3, band=band, verbose=verbose)
    wav_const = snrnorm.snr_constant_wav(
        wav, flux, snr=3, wav_ref=(lim[0] + lim[1]) / 2, verbose=verbose
    )
    assert band_const == 1
    assert wav_const == 1


@pytest.mark.parametrize(
    "wav,band",
    [
        (np.linspace(0.8, 1, 50), "VIS"),  # "VIS": (0.38, 0.78)
        (np.linspace(2, 3, 50), "J"),  # "J": (1.17, 1.33)
        (np.linspace(2.0, 2.1, 50), "K"),  # "K": (2.07, 2.35)
        (np.linspace(2.25, 2.4, 50), "K"),  # "K": (2.07, 2.35)
    ],
)
def test_snr_constant_band_with_invalid_wavelength(wav, band):
    with pytest.raises(ValueError):
        snrnorm.snr_constant_band(wav, np.ones(50), band=band)


@pytest.mark.parametrize("wav_ref", [0.5, 4])
def test_snr_constant_wav_ref_outside_wav(wav_ref):
    """Wav-ref outside bounds of wav should raise ValueError"""
    wav = np.linspace(1, 3, 60)
    flux = np.random.randn(len(wav))

    with pytest.raises(ValueError):
        snrnorm.snr_constant_wav(wav, flux, wav_ref)



import os

import pytest
import yaml

from eniric import DEFAULT_CONFIG_FILE, config
from eniric._config import Config

base_dir = os.path.dirname(__file__)
test_filename = os.path.join(base_dir, "data", "test_config.yaml")


class TestConfig:
    @pytest.fixture
    def test_config(self):
        """Config file for testing."""
        filename = test_filename
        yield Config(filename)

    def test_default_filename(self):
        default_config = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "eniric", "config.yaml"
        )
        assert DEFAULT_CONFIG_FILE == default_config

    def test_base_dots(self):
        assert config.paths == config["paths"]
        assert config.cache == config["cache"]
        assert config.atmmodel == config["atmmodel"]
        assert config.bands == config["bands"]
        assert config.custom_bands == config["custom_bands"]

    @pytest.mark.parametrize(
        "key, values",
        [
            ("phoenix_raw", ["..", "data", "phoenix-raw"]),
            ("btsettl_raw", ["..", "data", "btsettl-raw"]),
            ("atmmodel", ["..", "data", "atmmodel"]),
            ("precision_results", ["..", "data", "precision"]),
        ],
    )
    def test_default_paths_keys(self, key, values):
        assert config.paths[key] == os.path.join(*values)

    @pytest.mark.parametrize(
        "key, values",
        [
            ("phoenix_raw", ["phoenix-raw"]),
            ("btsettl_raw", ["btsettl-raw"]),
            ("atmmodel", ["..", "..", "data", "atmmodel"]),
            ("precision_results", ["..", "..", "data", "precision"]),
        ],
    )
    def test_paths_keys(self, test_config, key, values):
        assert test_config.paths[key] == os.path.join(*values)

    def test_paths(self):
        assert isinstance(config.paths, dict)

    def test_cache(self,):
        assert isinstance(config.cache, dict)
        assert config.cache["location"] == ".joblib"

    def test_atmmodel(self, test_config):
        assert isinstance(config.atmmodel, dict)
        assert config.atmmodel["base"] == "Average_TAPAS_2014"

    def test_default_bands(self):
        assert isinstance(config.bands, dict)
        assert isinstance(config.bands["all"], list)
        assert config.bands["all"] == [
            "VIS",
            "GAP",
            "Z",
            "Y",
            "J",
            "H",
            "K",
            "CONT",
            "NIR",
            "TEST",
        ]

    def test_bands(self, test_config):
        assert test_config.bands["all"] == ["K", "H", "J", "Y", "Z", "TEST"]

    def test_custom_bands(self):
        assert isinstance(config.custom_bands, dict)
        for value in config.custom_bands.values():
            assert isinstance(value, list)
            assert len(value) == 2

    def test_change_file(self):
        previous = config._path
        config.change_file(test_filename)
        assert config._path != previous
        assert config._path == test_filename
        config.change_file(previous)
        assert config._path == previous

    def test_set_attr_fail_on_default(self):
        with pytest.raises(RuntimeError):
            config.name = "Stephen King"

    def test_set_base_attr(self, test_config):
        previous_name = test_config.name
        test_config.name = "new name"
        assert test_config.name == "new name"
        test_config.name = previous_name
        assert test_config.name == previous_name

    def test_set_non_base_attr(self, test_config):
        old_path = test_config.paths["btsettl_raw"]
        test_config.paths["btsettl_raw"] = "testpath_btsettl_raw"
        assert test_config.paths["btsettl_raw"] == "testpath_btsettl_raw"
        test_config.paths["btsettl_raw"] = old_path
        assert test_config.paths["btsettl_raw"] == old_path

    @pytest.mark.parametrize("switch", [True, False])
    def test_copy_config(self, tmpdir, switch):
        previous_file = config._path
        assert not os.path.exists(tmpdir.join("config.yaml"))
        config.copy_file(tmpdir, switch=switch)
        if switch:
            assert str(tmpdir) in config.pathdir
        else:
            assert str(tmpdir) not in config.pathdir
        assert os.path.exists(tmpdir.join("config.yaml"))

        config.change_file(previous_file)  # Restore config

    def test_lazy_load(self, test_config):
        previous = test_config.cache["location"]
        base = test_config._config
        base["cache"].update({"location": "test_output"})
        with open(test_config._path, "w") as f:
            yaml.safe_dump(base, f)

        assert test_config.cache["location"] == "test_output"
        test_config.cache["location"] = previous
        assert test_config.cache["location"] == previous

    def test_pathdir(self):
        assert config.pathdir == os.path.split(config._path)[0]
        with pytest.raises(AttributeError):
            config.pathdir = 5

    def test_pathdir_getter(self):
        assert config.pathdir == config.get_pathdir()

    def test_update_config_with_None(self):
        with pytest.raises(
            RuntimeError, match="The default file is not allowed to be overwritten."
        ):
            # Default config protection
            config.update(d=None)

    def test_update_test_config_with_None(self, test_config):
        previous_config = test_config
        test_config.update(d=None)
        assert previous_config == test_config

    def test_update_test_config_with_dict(self, test_config):
        temp_name = "test name"
        temp_atmmodel = {"base": "test_Average_TAPAS"}
        previous_name = test_config.name
        previous_atmmodel = test_config.atmmodel
        assert previous_name != temp_name
        assert previous_atmmodel != temp_atmmodel
        test_config.update(d={"name": temp_name, "atmmodel": temp_atmmodel})
        assert test_config.name == temp_name
        assert test_config.atmmodel == temp_atmmodel
        test_config.update(d={"name": previous_name, "atmmodel": previous_atmmodel})
        assert test_config.atmmodel == previous_atmmodel
        assert test_config.name == previous_name

    def test_update_with_kwargs(self, test_config):
        temp_name = "test name"
        temp_atmmodel = {"base": "test_Average_TAPAS"}
        previous_name = test_config.name
        previous_atmmodel = test_config.atmmodel
        assert previous_name != temp_name
        assert previous_atmmodel != temp_atmmodel
        test_config.update(d=None, name=temp_name, atmmodel=temp_atmmodel)
        assert test_config.name == temp_name
        assert test_config.atmmodel == temp_atmmodel
        test_config.update(d={"name": previous_name}, atmmodel=previous_atmmodel)
        assert test_config.atmmodel == previous_atmmodel
        assert test_config.name == previous_name

    @pytest.mark.parametrize("key", ["name", "cache", "bands"])
    def test_delitem(self, test_config, key):
        # Ability to delete from config.
        __ = test_config[key]
        del test_config[key]
        with pytest.raises(KeyError):
            __ = test_config[key]



import numpy as np
import pytest
from hypothesis import given, settings, strategies as st
from joblib import Parallel

from eniric.broaden import (
    convolution,
    resolution_convolution,
    rotation_kernel,
    rotational_convolution,
    unitary_gaussian,
)


@settings(max_examples=100)
@given(
    st.lists(
        st.floats(
            min_value=1e-7, max_value=1e-5, allow_infinity=False, allow_nan=False
        ),
        unique=True,
        min_size=3,
        max_size=25,
    ),
    st.floats(min_value=1e-2, max_value=200),
    st.floats(min_value=1e-4, max_value=1),
)
def test_rotational_kernel(delta_lambdas, vsini, epsilon):
    """Test that the new and original code produces the same output."""
    delta_lambdas = np.sort(np.asarray(delta_lambdas), kind="quicksort")
    delta_lambdas = np.append(np.flipud(delta_lambdas), np.insert(delta_lambdas, 0, 0))
    delta_lambda_l = np.max(delta_lambdas) * 2

    new_profile = rotation_kernel(delta_lambdas, delta_lambda_l, vsini, epsilon)

    assert len(new_profile) == len(delta_lambdas)


@given(
    st.lists(
        st.floats(min_value=-100, max_value=100, allow_nan=False),
        min_size=1,
        unique=True,
    ),
    st.floats(min_value=-100, max_value=100, allow_nan=False),
    st.floats(min_value=0.001, max_value=100, allow_nan=False),
)
def test_unitary_gaussian(x, center, fwhm):
    """Just a quick simple test."""
    x = np.asarray(x)

    gaussian = unitary_gaussian(x, center, fwhm)
    print(gaussian)
    # point at center should be the max
    assert len(gaussian) == len(x)
    assert np.allclose(np.max(gaussian), gaussian[np.argmin(abs(x - center))])


def test_unitary_gaussian_type_errors():
    """Testing for type errors."""
    x = np.arange(-10, 10)
    center = 0
    fwhm = 3

    with pytest.raises(TypeError):
        unitary_gaussian(x, center, "fwhm")
    with pytest.raises(TypeError):
        unitary_gaussian(x, "center", fwhm)
    with pytest.raises(TypeError):
        unitary_gaussian(range(-10, 10), "center", fwhm)
    with pytest.raises(TypeError):
        unitary_gaussian(1, "center", fwhm)


@pytest.mark.parametrize("num_proc", [1])
def test_convolution_can_accept_int(num_proc):
    n = 5
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    convolution(x, y, vsini=1, R=1000, band="K", num_procs=num_proc)


@pytest.mark.parametrize("num_proc", [1])
def test_rot_convolution_can_accept_int(num_proc):
    n = 5
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    rotational_convolution(x, x, y, vsini=1, num_procs=num_proc)


@pytest.mark.parametrize("num_proc", [1])
def test_res_convolution_can_accept_int(num_proc):
    n = 5
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    resolution_convolution(x, x, y, R=1000, num_procs=num_proc)


@pytest.mark.parametrize("num_proc", [1, 2])
def test_convolution_can_accept_worker_pool(num_proc):
    n = 20
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    with Parallel(num_proc) as parallel:
        convolution(
            x, y, vsini=1, R=100_000, band="K", num_procs=parallel, verbose=False
        )


@pytest.mark.parametrize("num_proc", [1, 2])
def test_rot_convolution_can_accept_worker_pool(num_proc):
    n = 20
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    with Parallel(num_proc) as parallel:
        rotational_convolution(x, x, y, vsini=1, num_procs=parallel, verbose=False)


@pytest.mark.parametrize("num_proc", [1, 2])
def test_res_convolution_can_accept_worker_pool(num_proc):
    n = 20
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    with Parallel(num_proc) as parallel:
        resolution_convolution(x, x, y, R=100_000, num_procs=parallel, verbose=False)


@pytest.mark.parametrize("num_proc", [3.14, "str"])
def test_rot_convolution_with_bad_num_proc(num_proc):
    n = 20
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    with pytest.raises(
        TypeError, match="num_proc must be an int or joblib.parallel.Parallel"
    ):
        rotational_convolution(x, x, y, vsini=1, num_procs=num_proc, verbose=False)


@pytest.mark.parametrize("num_proc", [3.14, "str"])
def test_res_convolution_with_bad_num_proc(num_proc):
    n = 20
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    with pytest.raises(
        TypeError, match="num_proc must be an int or joblib.parallel.Parallel"
    ):
        resolution_convolution(x, x, y, R=100_000, num_procs=num_proc, verbose=False)


def test_convolution_can_accept_None():
    n = 20
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    convolution(x, y, vsini=1, R=100_000, band="K", num_procs=None)


def test_rot_convolution_can_accept_None():
    n = 20
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    rotational_convolution(x, x, y, vsini=1, num_procs=None)


def test_res_convolution_can_accept_None():
    n = 20
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    resolution_convolution(x, x, y, R=100_000, num_procs=None)


@pytest.mark.parametrize("zero_vsini", [0, 0.00, np.int(0.0), np.float64(0.0)])
def test_zero_rotation(zero_vsini):
    n = 20
    x = np.linspace(2.0, 2.3, n)
    y = np.random.randn(n)
    z = rotational_convolution(x, x, y, vsini=zero_vsini, num_procs=1)

    assert np.all(y == z)


@pytest.mark.parametrize("zero_vsini", [0, 0.00, np.int(0.0), np.float64(0.0)])
def test_zero_rotation_with_different_x(zero_vsini):
    n = 20
    x = np.linspace(1000, 2300, n)
    y = np.random.randn(n)
    z = rotational_convolution(x[4:16], x, y, vsini=zero_vsini, num_procs=1)

    assert np.all(y[4:16] == z)



"""Legacy functions kept for accessing changes to precision."""
from typing import Any, List, Optional, Tuple, Union

import numpy as np
from astropy import units as u
from astropy.units import Quantity
from numpy.core.multiarray import ndarray

from eniric.precision import rv_precision


# Legacy function in which the spectra is split into chunks first and then the pixel weights are calculated.
def RVprec_calc_masked(
    wavelength: Union[List[List[Any]], ndarray],
    flux: Union[ndarray, List[List[Any]]],
    mask: Optional[ndarray] = None,
    **kwargs,
) -> Quantity:
    """RV precision for split apart spectra.

    The same as rv_precision, but now wavelength and flux are organized into
    chunks according to the mask and the weighted average formula is used to
    calculate the combined precision.

    When considering the average RV as delivered by several slices of a
    spectrum, the error on the average is given by the error on a weighted
    average.
        mean(RV_rms) = 1 / sqrt(sum_i((1 / RV_rms(i))**2))

    Parameters
    ----------
    wavelength: array-like or list(array-like)
        Wavelength values of chunks.
    flux: array-like or list(array-like)
        Flux values of the chunks.
    mask: array-like of bool or None
        Mask of transmission cuts. Zero values are excluded and used to cut up
        the spectrum.
    kwargs:
        Kwargs for sqrt_sum_wis

    Returns
    -------
    RV_value: Quantity scalar
        Weighted average RV value of spectral chunks.

    Notes
    -----
    A "clump" is defined as a contiguous region of the array.
    Solution for clumping comes from
    https://stackoverflow.com/questions/14605734/numpy-split-1d-array-of-chunks-separated-by-nans-into-a-list-of-the-chunks
    """
    if mask is not None:
        # Turn wavelength and flux into masked arrays
        wavelength_clumps, flux_clumps = mask_clumping(wavelength, flux, mask)

    else:
        # When given a already clumped solution and no mask given.
        assert isinstance(wavelength, list)
        assert isinstance(flux, list)
        wavelength_clumps = wavelength
        flux_clumps = flux

    # Turn an ndarray into quantity array.
    # Need to use np.zeros instead of np.empty. Unassigned zeros are removed after with nonzero.
    # The "empty" values (1e-300) do not get removed and effect precision
    slice_rvs = Quantity(
        np.zeros(len(wavelength_clumps), dtype=float), unit=u.meter / u.second
    )  # Radial velocity of each slice

    for i, (wav_slice, flux_slice) in enumerate(zip(wavelength_clumps, flux_clumps)):
        if len(wav_slice) == 1:
            # Results in infinite rv, can not determine the slope of single point.
            continue

        else:
            wav_slice = np.asarray(wav_slice)
            flux_slice = np.asarray(flux_slice)
            slice_rvs[i] = rv_precision(wav_slice, flux_slice, **kwargs)

    # Zeros created from the initial empty array, when skipping single element chunks)
    slice_rvs = slice_rvs[np.nonzero(slice_rvs)]  # Only use nonzero values.
    rv_value = 1.0 / (np.sqrt(np.nansum((1.0 / slice_rvs) ** 2.0)))

    return rv_value


def mask_clumping(
    wave: ndarray, flux: ndarray, mask: ndarray
) -> Tuple[List[ndarray], List[ndarray]]:
    """Clump contiguous wavelength and flux sections into list.

    Note: Our value of mask (0 = bad points) is opposite to usage in
    np.ma.masked_array (1 = bad)
    Separate function to enable thorough testing.

    Parameters
    ----------
    wave: array-like of floats
        The wavelength array to clump.
    flux: array-like of floats
        The flux array to clump.
    mask: array-like of bool
        Boolean array with True indicating the values to use/keep.

    Returns
    -------
    wave_clumps: list(array)
        List of the valid wavelength sections.
    flux_clumps: list(array)
       List of the valid flux sections.

    """
    # Turn into masked array to use clump_unmasked method.
    mask = np.asarray(mask, dtype=bool)  # Make it bool so ~ works correctly

    masked_wave = np.ma.masked_array(wave, mask=~mask)  # ma mask is inverted
    masked_flux = np.ma.masked_array(flux, mask=~mask)  # ma mask is inverted

    wave_clumps = [wave[s] for s in np.ma.clump_unmasked(masked_wave)]
    flux_clumps = [flux[s] for s in np.ma.clump_unmasked(masked_flux)]

    return wave_clumps, flux_clumps


def RVprec_calc_weights_masked(
    wavelength: ndarray, flux: ndarray, mask: Optional[ndarray] = None, **kwargs
) -> Quantity:
    """RV precision setting weights of telluric lines to zero.

    Instead of splitting the spectra after every telluric line and
    individually calculating precision and taking the weighted average
    this just sets the pixel weights to zero.

    Parameters
    ----------
    wavelength: ndarray
        Wavelength array
    flux: ndarray
        Flux array
    mask: array or None
        Mask of transmission cuts. Zero values are excluded and used to cut up
        the spectrum.

    Returns
    -------
    RV_value: Quantity scalar
        RV precision.

    RVprec_calc_weights_masked is just a wrapper around rv_precision.

    """
    return rv_precision(wavelength, flux, mask=mask, **kwargs)



import numpy as np
import pytest
from astropy import units as u
from hypothesis import given, strategies as st

from eniric.atmosphere import Atmosphere
from eniric.legacy import RVprec_calc_masked, RVprec_calc_weights_masked, mask_clumping


def test_RV_prec_masked(test_spec):
    """Test same precision results between past pre-clumped version and mask clump version."""
    wav = test_spec[0]
    flux = test_spec[1]
    mask = test_spec[2]
    if mask is not None:
        mask = mask.round().astype(bool)
    else:
        mask = np.ones_like(wav)
    print(mask)

    # Pre clumping as done in Figueira et al. 2016
    wav_masked, flux_masked = mask_clumping(wav, flux, mask)
    rv_chunks = RVprec_calc_masked(wav_masked, flux_masked, mask=None)

    rv_masked = RVprec_calc_masked(wav, flux, mask)

    assert rv_masked.value == rv_chunks.value
    assert isinstance(rv_masked, u.Quantity)
    assert rv_masked.unit == u.m / u.s


@given(st.lists(st.booleans(), min_size=5, max_size=300))
def test_mask_clumping_of_mask(mask):
    """Masking mask show return all ones."""
    wav_clumped, flux_clumped = mask_clumping(mask, mask, mask)
    assert len(wav_clumped) == len(flux_clumped)
    for wav_i, flux_i in zip(wav_clumped, flux_clumped):
        assert np.all(wav_i)
        assert np.all(flux_i)
    # sum of masked_clumped should equal mask sum
    mask_sum = np.sum(mask)
    assert np.sum([np.sum(wav_i) for wav_i in wav_clumped]) == mask_sum
    assert np.sum([np.sum(flux_i) for flux_i in flux_clumped]) == mask_sum


def test_manual_clumping():
    """Test properties of clumping function using manually defined masked_arrays."""
    wav = np.arange(15)
    flux = np.arange(15, 30)
    mask = np.array([1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0])
    mask_bool = np.array([1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0], dtype=bool)

    wav_masked, flux_masked = mask_clumping(wav, flux, mask)
    wav_masked_bool, flux_masked_bool = mask_clumping(wav, flux, mask_bool)

    expected_wav = [np.arange(0, 4), np.arange(7, 10), np.arange(11, 14)]
    expected_flux = [np.arange(15, 19), np.arange(22, 25), np.arange(26, 29)]
    for i, __ in enumerate(wav_masked):
        assert np.allclose(wav_masked[i], expected_wav[i])
        assert np.allclose(flux_masked[i], expected_flux[i])
        assert np.allclose(wav_masked_bool[i], expected_wav[i])
        assert np.allclose(flux_masked_bool[i], expected_flux[i])

    assert len(expected_wav) == len(wav_masked)
    assert len(expected_flux) == len(flux_masked)


def test_legacy_RV_warns_nonfinite(grad_flag):
    """Some warning tests."""
    with pytest.warns(RuntimeWarning, match="divide by zero"):
        RVprec_calc_masked(
            np.array([1, 2, 3, 4]),
            np.array([1, 2, 3, 4]),
            np.array([0, 1, 0, 0]),
            grad=grad_flag,
        )


@pytest.mark.xfail(ModuleNotFoundError, reason="Issue with Starfish install.")
@pytest.mark.parametrize("band", ["H", "J", "K"])
def test_weights_clumping_grad(testing_spectrum, grad_flag, band):
    # Test masked clumping verse weight mask with new gradients
    # Test on an actual spectrum to check sizes of difference
    wav, flux = testing_spectrum
    atm = Atmosphere.from_band(band)
    mask = np.ones_like(wav)
    mask = atm.at(wav).mask

    clumped = RVprec_calc_masked(wav, flux, mask=mask, grad=grad_flag)
    weighted = RVprec_calc_weights_masked(wav, flux, mask=mask, grad=grad_flag)

    # They are not exactly the same by are within a specific percentage.
    ratio = (weighted.value - clumped.value) / clumped.value

    if grad_flag:
        # The difference for the new gradient is smaller.
        assert abs(ratio) < 0.008
    else:
        assert abs(ratio) < 0.04
    assert clumped.unit == weighted.unit



#!/usr/bin/env python

"""
untar_here.py
-------------
Bundled script to un-tar the eniric data downloaded.

Uses the tarfile module to extract the data.

"""
import argparse
import sys
import tarfile


def _parser():
    """Take care of all the argparse stuff."""
    parser = argparse.ArgumentParser(description="Extract from a tar file.")

    parser.add_argument("filename", help="File to untar.", type=str, default="")
    return parser.parse_args()


if __name__ == "__main__":
    filename = _parser().filename

    with tarfile.open(filename, "r") as tar:
        tar.extractall()
    sys.exit(0)



__version__ = "1.0rc1"

import os
import warnings

eniric_dir = os.path.dirname(__file__)
DEFAULT_CONFIG_FILE = os.path.join(eniric_dir, "config.yaml")

from ._config import Config

if (
    os.path.exists("config.yaml")
    and os.path.abspath("config.yaml") != DEFAULT_CONFIG_FILE
):
    config = Config("config.yaml")
else:
    warnings.warn(
        "Using the default config.yaml file located at {0}. This is likely NOT what you want and "
        "you will not be able to change any of the config values. Please use config.copy_file(<path>) to copy a "
        "version of the default config for your own project.".format(
            DEFAULT_CONFIG_FILE
        ),
        UserWarning,
    )
    config = Config(DEFAULT_CONFIG_FILE)

__all__ = [
    "atmosphere",
    "broaden",
    "config",
    "corrections",
    "legacy",
    "io_module",
    "precision",
    "resample",
    "snr_normalization",
    "utilities",
]



"""To test the equivalence of old and new code to check if it does the same thing."""

import numpy as np
import pytest

import eniric.io_module as io
import eniric.utilities as utils


def test_write_read_2col(tmpdir):
    """Check pd_writer same write_col with with exponential flag."""
    data = np.random.randn(2, 100) * 1e7

    tmp_file = tmpdir.join("file_col2.dat")
    tmp_file_name = str(tmp_file)
    # write files
    io.pdwrite_2col(tmp_file_name, data[0], data[1])

    # re-read files
    a = io.pdread_2col(tmp_file_name)

    # check results the same
    assert np.allclose(a[0], data[0])
    assert np.allclose(a[1], data[1])

    # clean-up
    utils.silent_remove(tmp_file_name)


def test_write_read_3col(tmpdir):
    """Check pd_writer same write_col with with exponential flag."""
    data = np.random.randn(3, 100) * 1e7

    tmp_file = tmpdir.join("file_col3.dat")
    tmp_file_name = str(tmp_file)

    # write files
    io.pdwrite_3col(tmp_file_name, data[0], data[1], data[2])

    # re-read files
    a = io.pdread_3col(tmp_file_name)

    # check results the same
    assert np.allclose(a[0], data[0])
    assert np.allclose(a[1], data[1])
    assert np.allclose(a[2], data[2])

    # clean-up
    utils.silent_remove(tmp_file_name)


def test_pdwrire_cols(tmpdir):
    """Test writer that can take variable column numbers."""
    tmpfile = tmpdir.join("pd_multicol_test.dat")
    pd_multicol_name = str(tmpfile)

    data1 = range(5)
    data2 = range(5, 10)

    # 0 means successful write
    assert 0 == io.pdwrite_cols(pd_multicol_name, data1, data2, data1)
    assert 0 == io.pdwrite_cols(pd_multicol_name, data1)
    assert 0 == io.pdwrite_cols(
        pd_multicol_name,
        data1,
        data2,
        data1,
        data2,
        header=["headers", "for", "column", "labels"],
    )
    assert 0 == io.pdwrite_cols(pd_multicol_name, data1, data2, sep=",", index=True)
    # clean-up
    utils.silent_remove(pd_multicol_name)


def test_write_col_errors_different_length(tmpdir):
    tmpfile = tmpdir.join("pd_multicol_test.dat")
    pd_multicol_name = str(tmpfile)

    data1 = range(5)
    data2 = range(5, 10)
    bad_data = range(6)  # Different length

    # test uneven data lengths
    with pytest.raises(ValueError):
        io.pdwrite_cols(pd_multicol_name, data1, data2, bad_data)


def test_write_col_errors(tmpdir):
    tmpfile = tmpdir.join("pd_multicol_test.dat")
    pd_multicol_name = str(tmpfile)

    data1 = range(5)
    data2 = range(5, 10)
    bad_data = range(6)  # Different length

    # test bad header
    with pytest.raises(ValueError):
        io.pdwrite_cols(
            pd_multicol_name, data1, data2, bad_data, header=["too", "many", "values"]
        )


def test_write_col_errors_bad_keyword(tmpdir):
    tmpfile = tmpdir.join("pd_multicol_test.dat")
    pd_multicol_name = str(tmpfile)

    data1 = range(5)

    with pytest.raises(TypeError):
        io.pdwrite_cols(pd_multicol_name, data1, bad="keyword")



import os
from os.path import join

import astropy.units as u
import numpy as np
import pandas as pd
import pytest

import eniric.io_module as io
from eniric import config
from eniric.atmosphere import Atmosphere
from eniric.utilities import load_aces_spectrum


@pytest.fixture
def published_data():
    name = "data/precision/precision_figueira_2016.dat"
    df = pd.read_csv(name, sep="\t")
    return df


@pytest.fixture(
    params=[
        ("M0", "Z", 1, "60k"),
        ("M0", "J", 1, "100k"),
        ("M3", "Y", 5, "80k"),
        ("M6", "H", 10, "100k"),
        ("M9", "K", 1, "80k"),
    ]
)
def model_parameters(request):
    # Tuple of "SpType, band, vsini, R"
    return request.param


# Define some fixtures for eniric.precision.
per_s_cm2 = (1 / u.second) / (u.centimeter ** 2)


@pytest.fixture(
    params=[
        (np.arange(1, 101), np.random.random(100), None),
        (np.linspace(2.1, 2.5, 200), np.random.random(200), np.random.random(200)),
        (
            np.linspace(0.5, 1.5, 50),
            np.random.random(50),
            np.floor(2 * np.random.random(50)),
        ),
    ]
)
def test_spec(request):
    """Wave and flux, mask examples."""
    wav, flux, mask = request.param

    # Modify mask so that there are at least 2 consecutive zeros in "random" mask.
    if mask is not None:
        new_mask = np.ones_like(mask)
        new_mask[:-1] = new_mask[:-1] + mask[1:]  # Left add
        new_mask[1:] = new_mask[1:] + mask[:-1]  # Right add
        new_mask = new_mask > 2
        return wav, flux, new_mask.astype(int)
    else:
        return wav, flux, mask


# 3 situations each variable, no unit. a unit. or a dimensionless unscaled unit.
@pytest.fixture(params=[1, u.micron, u.dimensionless_unscaled])
def wav_unit(request):
    """Iterate some units on wavelength."""
    return request.param


@pytest.fixture(params=[1, per_s_cm2, u.dimensionless_unscaled])
def flux_unit(request):
    """Iterate some units on flux"""
    return request.param


@pytest.fixture(params=[1, u.dimensionless_unscaled])
def trans_unit(request):
    """Iterate some units on mask/transmission."""
    return request.param


@pytest.fixture(params=[True, False])
def grad_flag(request):
    """Gradient flag parameter."""
    return request.param


@pytest.fixture(
    params=[
        # "Average_TAPAS_2014_H.dat",
        "Average_TAPAS_2014_K.dat",
        "Average_TAPAS_2014_J.dat",
    ]
)
def atm_model(request):
    """Get atmospheric model name to load."""
    return join(config.pathdir, config.paths["atmmodel"], request.param)


@pytest.fixture(params=[2.5, 4])
def atmosphere_fixture(request, atm_model):
    percent_cutoff = request.param
    atm = Atmosphere.from_file(atm_model)
    atm.mask_transmission(percent_cutoff)
    return atm


@pytest.fixture()
def short_atmosphere(atmosphere_fixture):
    # First 2000 data points only to speed up tests
    return atmosphere_fixture[:2000]


@pytest.fixture(params=[(0, 1500), (8000, 9000)])
def sliced_atmmodel_default_mask(request, atm_model):
    """To do own masking. Sliced in different places."""
    lower, upper = request.param  # slice limits
    atm = Atmosphere.from_file(atm_model)
    return atm[int(lower) : int(upper)]


@pytest.fixture(params=[[3900, 4.5, 0, 0], [2600, 4.5, 0, 0]])
def testing_spectrum(request, use_test_config):
    wav, flux = load_aces_spectrum(request.param)
    return wav, flux


@pytest.fixture(scope="function")
def use_test_config():
    """Change configuration used to the test_config file."""
    original = config._path
    base_dir = os.path.dirname(__file__)
    test_filename = os.path.join(base_dir, "tests", "data", "test_config.yaml")
    config.change_file(test_filename)
    yield None
    config.change_file(original)



"""
The theoretical RV precision (only due to photon noise) is
inversely dependent on the spectral flux.
That is as the flux increases the error on the RV measurement decreases.

The flux of a observed spectra is dependant on a number of
factors e.g.
    * Star luminosity
    * Integration time
    * Telescope area
    * Instrument efficiency

The photon flux can be scaled by multiplicative constant which
affects the photon SNR and the radial velocity precision.

To compare the relative theoretical precision between different
synthetic spectra they are normalized consistently.

This is achieved by normalizing to a specific SNR per
resolution element at a particular location.

By default this is a SNR of 100 at the center of the "J" band.
Other band centers or wavelengths as well as other SNR values
can be chosen.

The following functions are used to determine the normalization
constant to divide a spectrum by to achieve the desired
normalization.

"""

import math
from typing import Optional, Union

import numpy as np
from numpy import ndarray

import eniric.utilities as utils


def snr_constant_band(
    wav: ndarray,
    flux: ndarray,
    snr: Union[int, float] = 100,
    band: str = "J",
    sampling: Union[int, float] = 3.0,
    verbose: bool = False,
) -> float:
    """Determine the normalization constant to achieve a SNR in the middle of a given band.

    SNR estimated by the square root of the number of photons in a resolution element.

    Parameters
    ----------
    wav: ndarray
        Wavelength array in microns.
    flux: ndarray
        Photon flux array (photons/s/cm**2).
    snr: int or float
        SNR per resolution element to achieve. Default is 100.
    band: str
        Band to use for normalization. Default is "J".
    sampling: int or float
       Number of pixels per resolution element. Default is 3.
    verbose: bool
        Enable verbose. Default is False.

    Returns
    -------
    norm_value: float
        Normalization value to divide spectrum by to achieve a
        signal-to-noise level of snr within an resolution element
        in the middle of the band.

    Note
    ----
    This is a wrapper around `snr_constant_wav`, using the band center.

    Warning
    -------
    Wavelength is expected in microns!

    """
    band_middle = utils.band_middle(band)

    if not (wav[0] < band_middle < wav[-1]):
        raise ValueError("Band center not in wavelength range.")

    norm_value = snr_constant_wav(
        wav, flux, wav_ref=band_middle, snr=snr, sampling=sampling, verbose=verbose
    )

    return norm_value


def snr_constant_wav(
    wav: ndarray,
    flux: ndarray,
    wav_ref: float,
    snr: Union[int, float] = 100,
    sampling: Union[int, float] = 3.0,
    verbose: bool = False,
) -> float:
    """Determine the normalization constant to achieve a SNR at given wavelength.

    SNR estimated by the square root of the number of photons in a resolution element.

    Parameters
    ----------
    wav: ndarray
        Wavelength array.
    flux: ndarray
        Photon flux array (photons/s/cm**2).
    wav_ref: float
        Wavelength to set the SNR per resolution element.
    snr: int, float
        SNR per resolution element to achieve. Default is 100.
    sampling: int or float
       Number of pixels per resolution element. Default is 3.
    verbose: bool
        Enable verbose. Default is False.

    Returns
    -------
    norm_value: float
        Normalization value to divide the flux by to achieve the desired SNR "SNR"
        in resolution element (defined by "sampling") around the wavelength "wav_ref".

    Note
    ----
    If sampling is a float it will rounded to the nearest integer for indexing.

    """
    if wav_ref < wav[0] or wav_ref > wav[-1]:
        raise ValueError("Reference wavelength is outside of the wavelength bounds")
    index_ref = np.searchsorted(wav, [wav_ref])[0]  # Searching for the closest index

    indexes = sampling_index(
        index_ref, sampling=np.round(sampling), array_length=len(wav)
    )

    snr_estimate = np.sqrt(np.sum(flux[indexes]))

    if verbose:
        print(
            "\tSanity Check: The reference S/N at {1:3.02f} was of {0:4.2f}.".format(
                snr_estimate, wav_ref
            )
        )
    norm_value = (snr_estimate / snr) ** 2
    return norm_value


def sampling_index(
    index: int, sampling: int = 3, array_length: Optional[int] = None
) -> ndarray:
    """Get a small number of index values around the given index value.

    Parameters
    ----------
    index: int
        The index value which to select values around.
    sampling: int
        Number of index values to return (sampling per resolution element). Must be an integer. Default is 3.
    array_length: int or None
        Length of array the indexes will be used in. To not exceed array length. Default is None.

    Returns
    -------
    indexes: ndarray
        The index values.

    """

    half_sampling = math.floor(sampling / 2)
    if sampling % 2 == 0:  # even sampling
        # index values must be integer
        indexes = np.arange(index - half_sampling, index + half_sampling, dtype=int)
        assert len(indexes) % 2 == 0  # confirm even
        assert len(indexes) == sampling
    else:
        indexes = index + np.arange(-half_sampling, sampling - half_sampling, dtype=int)
        assert len(indexes) % 2 != 0  # confirm odd
        assert len(indexes) == sampling

    if array_length is not None:
        if np.any(indexes >= array_length):
            # This may need fixed up in the future.
            raise ValueError("Indexes has values greater than the length of array.")

    if np.any(indexes < 0):
        raise ValueError("Indexes has values less than 0.")

    return indexes



"""Test of atmosphere.py functions."""

import numpy as np
import pytest
from hypothesis import assume, given, strategies as st

from eniric.atmosphere import Atmosphere, consecutive_truths
from eniric.broaden import resolution_convolution
from eniric.utilities import band_limits

size = 50  # Size of arrays if need consistent length


@pytest.mark.xfail()
def test_Atmosphere_funtional_test(short_atmosphere):
    """ Check Mask still works after shifting and masks out correct values.
    Load in telluric file.
    Mask the lines,
    Bary shift the lines.
    Check masks are still ok.
    """
    atm = short_atmosphere
    atm.verbose = True
    atm.mask = np.ones_like(atm.wl, dtype=bool)  # reset mask
    assert not np.all(atm.transmission[atm.mask] >= 0.98)
    atm.broaden(100_000)
    atm.mask_transmission(2)
    assert np.all(atm.transmission[atm.mask] >= 0.98)
    atm.barycenter_broaden(consecutive_test=True)
    assert np.all(atm.transmission[atm.mask] >= 0.98)


@pytest.mark.parametrize(
    "wave, transmission, std, mask",
    [
        ([1, 2, 3, 4], [0.5, 0.6, 0.7, 8], [0.0, 0.2, 0.1, 0.2], [0, 1, 1, 1]),
        ([2000, 2100], [0.97, 0.99], [0.1, 0.1], [False, True]),
        ([], [], [], []),
    ],
)
def test_atmosphere_class(wave, transmission, std, mask):
    atmos = Atmosphere(
        np.array(wave), np.array(transmission), np.array(mask), std=np.array(std)
    )
    assert np.all(atmos.wl == wave)
    assert np.all(atmos.transmission == transmission)
    assert np.all(atmos.mask == mask)
    assert np.all(atmos.std == std)
    assert atmos.mask.dtype == np.bool


@pytest.mark.parametrize(
    "wave, transmission, std, mask ", [([1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12])]
)
def test_atmosphere_class_turns_lists_to_arrays(wave, transmission, std, mask):
    atmos = Atmosphere(wave, transmission, mask, std)
    assert isinstance(atmos.wl, np.ndarray)
    assert isinstance(atmos.transmission, np.ndarray)
    assert isinstance(atmos.std, np.ndarray)
    assert isinstance(atmos.mask, np.ndarray)
    assert atmos.mask.dtype == np.bool


@pytest.mark.parametrize(
    "wave, transmission",
    [([1, 2, 3], [0.4, 0.5, 0.6]), ([7, 8, 9], [0.10, 0.11, 0.12])],
)
def test_atmosphere_class_nomask(wave, transmission):
    atmos = Atmosphere(wave, transmission)
    assert np.all(atmos.wl == wave)
    assert np.all(atmos.transmission == transmission)
    assert np.all(atmos.mask == 1)
    assert len(atmos.mask) == len(atmos.transmission)
    assert atmos.mask.dtype == np.bool


def test_atmosphere_from_file(atm_model):
    atmos = Atmosphere.from_file(atmmodel=atm_model)
    assert len(atmos.wl) == len(atmos.transmission)
    assert len(atmos.transmission[atmos.mask]) != len(
        atmos.transmission
    )  # mask is not all ones


@given(
    trans=st.lists(st.floats(min_value=0, max_value=1), min_size=size, max_size=size),
    percent=st.floats(min_value=0.5, max_value=99),
)
def test_atmosphere_masking(trans, percent):
    cutoff = 1 - (percent / 100.0)

    assume(np.any(np.array(trans) < cutoff))
    assume(np.any(np.array(trans) >= cutoff))

    atmos = Atmosphere(wavelength=np.arange(len(trans)), transmission=trans)
    org_mask = atmos.mask
    atmos.mask_transmission(percent)

    assert np.any(atmos.mask != org_mask)
    assert np.all(atmos.transmission[atmos.mask] >= cutoff)
    assert np.all(atmos.transmission[~atmos.mask] < cutoff)


@pytest.mark.parametrize("rv", [-1471, 65])  # some fixed RV values
def test_values_within_the_rv_of_telluric_lines_are_masked(
    sliced_atmmodel_default_mask, rv
):
    # Enable object
    atmos = sliced_atmmodel_default_mask
    org_mask = atmos.mask.copy()
    # RV shift mask
    atmos.barycenter_broaden(rv=rv, consecutive_test=False)

    for pixel, mask_value, org_val in zip(atmos.wl, atmos.mask, org_mask):
        if mask_value != 0:
            assert org_val != 0
            # Find rv limits to this pixel.
            wl_lower, wl_upper = pixel * (1 - rv / 3e5), pixel * (1 + rv / 3e5)
            wl_mask = (atmos.wl >= wl_lower) * (atmos.wl < wl_upper)
            assert np.all(atmos.mask[wl_mask] == 1)


@pytest.mark.parametrize("consec_test", [True, False])
def test_atmos_barycenter_shift_mask(sliced_atmmodel_default_mask, consec_test):
    """Test barycentric shift code."""
    # Bary mask should have more pixels mask (at 0) so count will be lower
    atmos = sliced_atmmodel_default_mask
    org_mask = atmos.mask.copy()
    org_number_masked = np.sum(org_mask)
    org_len = len(org_mask)
    atmos.barycenter_broaden(consecutive_test=consec_test)
    new_number_masked = np.sum(atmos.mask)

    assert (new_number_masked < org_number_masked) or (
        (org_len == np.sum(org_number_masked)) or (org_number_masked == 0)
    )
    assert len(atmos.mask) == org_len
    assert np.all(
        (atmos.mask * org_mask) == atmos.mask
    )  # zeros should not turn into ones


def test_consecutive_truths():
    """Test consecutive truths lists count of consecutive ones."""
    array1 = np.array([1, 1, 1, 0, 1, 1, 0, 0, 1, 0], dtype=bool)  # [3,2 1]
    array2 = np.array([0, 0, 1, 1, 1, 1, 1, 0, 1, 1], dtype=bool)  # [5, 2]
    array3 = np.array([1, 0, 1, 0, 1, 0, 1, 0, 1, 0], dtype=bool)  # [1,1,1,1,1]
    array4 = np.zeros(5, dtype=bool)  # []
    array5 = np.ones(5, dtype=bool)  # [5]

    assert np.all(consecutive_truths(array1) == [3, 2, 1])
    assert np.all(consecutive_truths(array2) == [5, 2])
    assert np.all(consecutive_truths(array3) == [1, 1, 1, 1, 1])
    assert np.all(consecutive_truths(array5) == [5])
    assert np.all(consecutive_truths(array4) == [])

    # Check sum of trues equal sum of returned list
    assert np.sum(array1) == np.sum(consecutive_truths(array1))

    rand_array = np.asarray(
        np.floor(np.random.random(50) * 2), dtype=bool
    )  # random values
    assert np.sum(rand_array) == np.sum(consecutive_truths(rand_array))


@pytest.mark.xfail()
@pytest.mark.parametrize("resolution", [1000, 50000])
def test_atmos_broadening(atmosphere_fixture, resolution):
    # Test broadening transmission preforms instrumental convolution
    atm = atmosphere_fixture[:4000]
    atm_org = atm.copy()

    #    normalize: bool = True,

    new_trans = resolution_convolution(
        atm_org.wl,
        atm_org.wl,
        atm_org.transmission,
        R=resolution,
        fwhm_lim=5,
        num_procs=1,
        normalize=True,
    )

    atm.broaden(resolution=resolution, num_procs=1)

    assert not np.allclose(
        atm.transmission, atm_org.transmission
    )  # Should have changed
    assert np.allclose(atm.transmission, new_trans)  # Should be equal

    assert np.allclose(atm.wl, atm_org.wl)  # Should not have changed
    assert np.allclose(atm.mask, atm_org.mask)  # Should not have changed
    assert np.allclose(atm.std, atm_org.std)  # Should not have changed


@pytest.mark.xfail()
@pytest.mark.parametrize("resolution", [1000, 20000])
def test_atmos_broadening_reduces_number_of_masked_points(
    atmosphere_fixture, resolution
):
    """Test broadening transmission preforms instrumental convolution."""
    percent = 4
    cuttoff = 1 - percent / 100
    atm = atmosphere_fixture[:4000]
    atm.mask_transmission(percent)

    atm_org = atm.copy()

    atm.broaden(resolution=resolution, num_procs=1)
    assert np.allclose(atm.wl, atm_org.wl)  # Should not have changed
    assert np.allclose(atm.mask, atm_org.mask)  # Should not have changed
    assert np.allclose(atm.std, atm_org.std)  # Should not have changed

    # Transmission changes but number of points in mask may not
    assert not np.allclose(
        atm.transmission, atm_org.transmission
    )  # Should have changed
    atm.mask_transmission(percent)

    if np.allclose(atm.mask, atm_org.mask):  # Mask changes after convolution
        # If masks don't change check both conditions still the same
        assert np.all((atm.transmission > cuttoff) == (atm_org.transmission > cuttoff))


def test_Atmosphere_has_getitem():
    assert hasattr(Atmosphere, "__getitem__")


def test_Atmosphere_sliceable(short_atmosphere):
    atm = short_atmosphere
    assert len(atm.wl) != 100
    # Indexing on object.
    atm2 = atm[:100]

    assert len(atm2.wl) == 100
    assert len(atm2.transmission) == 100
    assert len(atm2.std) == 100
    assert len(atm2.mask) == 100


def test_Atmosphere_copyable(short_atmosphere):
    atm = short_atmosphere[10:50]

    atm2 = atm.copy()

    assert np.all(atm.wl == atm2.wl)
    assert not (atm.wl is atm2.wl)

    assert np.all(atm.transmission == atm2.transmission)
    assert not (atm.transmission is atm2.transmission)
    assert np.all(atm.std == atm2.std)
    assert not (atm.std is atm2.std)
    assert np.all(atm.mask == atm2.mask)
    assert not (atm.mask is atm2.mask)


def test_Atmosphere_copyable_attr():
    assert hasattr(Atmosphere, "copy")


def test_Atmosphere_band_select():
    """Small test to check band selection."""
    band = "K"  # "K": (2.07, 2.35),

    atm = Atmosphere(
        [2.0, 2.1, 2.2, 2.3, 2.4, 2.5], np.arange(6), std=None, mask=[1, 0, 0, 1, 1, 0]
    )
    atm2 = atm.band_select(band)

    assert np.allclose(atm2.wl, [2.1, 2.2, 2.3])
    assert np.allclose(atm2.mask, [0, 0, 1])


@pytest.mark.parametrize("band", ["TEST", "K"])
def test_from_band_is_a_constructor(band):
    """This is testing loading files using from_band method."""
    atm = Atmosphere.from_band(band)

    min_wl, max_wl = band_limits(band)
    assert isinstance(atm, Atmosphere)
    assert np.all(atm.wl <= max_wl * 1.01)
    assert np.all(atm.wl >= min_wl * 0.99)



import pytest

from eniric.corrections import correct_artigau_2018


@pytest.mark.parametrize("band", ["g", "Z", "z", "K", "H", "J", "Y"])
def test_bands_in_correction(band):
    value = correct_artigau_2018(band)
    assert isinstance(value, float)
    assert value > 0.2
    assert value < 1.5


@pytest.mark.parametrize("band", ["x", "p", "a"])
def test_bands_not_in_correction(band):
    with pytest.raises(KeyError):
        correct_artigau_2018(band)


@pytest.mark.parametrize("band, value", [("Y", 0.29), ("K", 1.47)])
def test_corection_specific_values(band, value):
    assert correct_artigau_2018(band) == value



#!/usr/bin/env python
import argparse
import sys
from itertools import product
from os.path import join

import numpy as np

from eniric import config

try:
    from Starfish.grid_tools import download_PHOENIX_models
except ImportError:
    raise ImportError("You need to install Starfish.")


def _parser():
    """Take care of all the argparse stuff."""
    parser = argparse.ArgumentParser(
        description="Download Phoenix ACES test spectra. A single spectra each for an M0, M3, M6, and M9 M-dwarf."
    )

    parser.add_argument(
        "-o",
        "--outdir",
        help="Output directory. Default is configured by the config.yaml.",
        type=str,
        default=join(config.pathdir, config.paths["phoenix_raw"]),
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = _parser()
    outdir = args.outdir

    models = np.array(list(product((2600, 2800, 3500, 3900), (4.5,), (0.0,))))
    download_PHOENIX_models(models, outdir)
    sys.exit(0)



#!/usr/bin/env python
"""
phoenix_precision.py
--------------------
Script to generate RV precision of synthetic spectra, see :ref:`Calculating-Precisions`.

"""
import argparse
import itertools
import os
import warnings
from typing import Any, List, Optional, Tuple

import numpy as np
from astropy import units as u
from astropy.units import Quantity
from numpy import ndarray

from eniric import config
from eniric.atmosphere import Atmosphere
from eniric.broaden import convolution
from eniric.corrections import correct_artigau_2018
from eniric.precision import quality, rv_precision
from eniric.resample import log_resample
from eniric.snr_normalization import snr_constant_band
from eniric.utilities import (
    band_middle,
    cpu_minus_one,
    doppler_shift_flux,
    load_aces_spectrum,
    load_btsettl_spectrum,
)

num_cpu_minus_1 = cpu_minus_one()

ref_choices = ["SELF"]
ref_choices.extend(config.bands["all"])


def _parser():
    """Take care of all the argparse stuff."""
    parser = argparse.ArgumentParser(
        description="Calculate precision and quality for synthetic spectra."
    )
    parser.add_argument(
        "-t",
        "--temp",
        type=int,
        default=[3900],
        help="Temperature, default=[3900].",
        nargs="+",
    )
    parser.add_argument(
        "-l",
        "--logg",
        type=float,
        default=[4.5],
        help="Logg, default = [4.5].",
        nargs="+",
    )
    parser.add_argument(
        "-m",
        "--metal",
        type=float,
        default=[0.0],
        help="Metallicity, default=[0.0].",
        nargs="+",
    )
    parser.add_argument(
        "-a",
        "--alpha",
        type=float,
        default=[0.0],
        help="Alpha, default=[0.0].",
        nargs="+",
    )
    parser.add_argument(
        "-s",
        "--sampling",
        type=float,
        default=[3.0],
        help="Sampling, default=[3.0].",
        nargs="+",
    )
    parser.add_argument(
        "-r",
        "--resolution",
        type=float,
        default=[50000],
        help="Instrumental resolution, default=[50000]",
        nargs="+",
    )
    parser.add_argument(
        "-v",
        "--vsini",
        type=float,
        default=[1.0],
        help="Rotational Velocity, default = [1.0]",
        nargs="+",
    )
    parser.add_argument(
        "-b",
        "--bands",
        type=str,
        default=["J"],
        choices=config.bands["all"],
        help="Wavelength bands to select, default=['J'].",
        nargs="+",
    )
    parser.add_argument(
        "--model",
        type=str,
        default="aces",
        choices=["aces", "btsettl", "phoenix"],
        help="Spectral models to use, default='aces'.",
    )
    parser.add_argument(
        "--snr", help="Mid-band SNR scaling, default=100.", default=100, type=float
    )
    parser.add_argument(
        "--ref_band",
        help="SNR reference band, default='J'. "
        "'self' scales each band relative to the SNR itself.",
        choices=ref_choices,
        default="J",
        type=str,
    )
    parser.add_argument(
        "--num_procs",
        help="Number of processors to use, default = (Total cores - 1)",
        default=num_cpu_minus_1,
        type=int,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Filename for result file, default='precisions.csv'.",
        default="precisions.csv",
        type=str,
    )
    parser.add_argument(
        "--rv",
        help="Radial velocity value, default=[0.0]",
        default=[0.0],
        type=float,
        nargs="+",
    )
    parser.add_argument(
        "--add_rv", help="Include a radial velocity shift.", action="store_true"
    )
    parser.add_argument(
        "--air", help="Convert to air wavelengths.", action="store_true"
    )
    parser.add_argument(
        "--correct",
        help="Apply Artigau et al. (2018) RV band corrections.",
        action="store_true",
    )

    parser.add_argument("-V", "--verbose", help="Enable verbose.", action="store_true")
    parser.add_argument(
        "--disable_normalization",
        help="Disable the convolution normalization.",
        action="store_true",
    )
    return parser.parse_args()


def do_analysis(
    star_params,
    vsini: float,
    R: float,
    band: str,
    sampling: float = 3.0,
    conv_kwargs=None,
    snr: float = 100.0,
    ref_band: str = "J",
    rv: float = 0.0,
    air: bool = False,
    model: str = "aces",
    verbose: bool = False,
) -> Tuple[Quantity, ...]:
    """Calculate RV precision and Quality for specific parameter set.

    Parameters
    ----------
    star_param:
        Stellar parameters [temp, logg, feh, alpha] for phoenix model libraries.
    vsini: float
       Stellar equatorial rotation.
    R: float
        Instrumental resolution.
    band: str
        Spectral band.
    sampling: float (default=False)
        Per pixel sampling (after convolutions)
    conv_kwargs: Dict (default=None)
        Arguments specific for the convolutions,
        'epsilon', 'fwhm_lim', 'num_procs', 'normalize', 'verbose'.
    snr: float (default=100)
        SNR normalization level. SNR per pixel and the center of the ref_band.
    ref_band: str (default="J")
        Reference band for SNR normalization.
    rv: float
        Radial velocity in km/s (default = 0.0).
    air: bool
        Get model in air wavelengths (default=False).
    model: str
        Name of synthetic library (aces, btsettl) to use. Default = 'aces'.
    verbose:
        Enable verbose (default=False).

    Returns
    -------
    q: astropy.Quality
     Spectral quality.
    result_1: astropy.Quality
        RV precision under condition 1.
    result_2 : astropy.Quality
        RV precision under condition 2.
    result_3: astropy.Quality
        RV precision under condition 3.

    Notes
    -----
        We apply the radial velocity doppler shift after
            - convolution (rotation and resolution)
            - resampling
            - SNR normalization.

        in this way the RV only effects the precision due to the telluric mask interaction.
        Physically the RV should be applied between the rotational and instrumental convolution
        but we assume this effect is negligible.

    """
    if conv_kwargs is None:
        conv_kwargs = {
            "epsilon": 0.6,
            "fwhm_lim": 5.0,
            "num_procs": num_cpu_minus_1,
            "normalize": True,
            "verbose": verbose,
        }

    if ref_band.upper() == "SELF":
        ref_band = band

    model = check_model(model)

    if model == "aces":
        wav, flux = load_aces_spectrum(star_params, photons=True, air=air)
    elif model == "btsettl":
        wav, flux = load_btsettl_spectrum(star_params, photons=True, air=air)
    else:
        raise Exception("Invalid model name reached.")

    wav_grid, sampled_flux = convolve_and_resample(
        wav, flux, vsini, R, band, sampling, **conv_kwargs
    )

    # Doppler shift
    try:
        if rv != 0:
            sampled_flux = doppler_shift_flux(wav_grid, sampled_flux, vel=rv)
    except Exception as e:
        print("Doppler shift was unsuccessful")
        raise e

    # Scale normalization for precision
    wav_ref, sampled_ref = convolve_and_resample(
        wav, flux, vsini, R, ref_band, sampling, **conv_kwargs
    )
    snr_normalize = snr_constant_band(
        wav_ref, sampled_ref, snr=snr, band=ref_band, sampling=sampling, verbose=verbose
    )
    sampled_flux = sampled_flux / snr_normalize

    if (ref_band == band) and verbose:
        mid_point = band_middle(ref_band)
        index_ref = np.searchsorted(
            wav_grid, mid_point
        )  # searching for the index closer to 1.25 micron
        snr_estimate = np.sqrt(np.sum(sampled_flux[index_ref - 1 : index_ref + 2]))
        print(
            "\tSanity Check: The S/N at {0:4.02} micron = {1:4.2f}, (should be {2:g}).".format(
                mid_point, snr_estimate, snr
            )
        )

    # Load Atmosphere for this band.
    atm = Atmosphere.from_band(band=band, bary=True).at(wav_grid)
    assert np.allclose(atm.wl, wav_grid), "The atmosphere does not cover the wav_grid"

    # Spectral Quality/Precision
    q = quality(wav_grid, sampled_flux)

    # Precision given by the first condition:
    result_1 = rv_precision(wav_grid, sampled_flux, mask=None)

    # Precision as given by the second condition
    result_2 = rv_precision(wav_grid, sampled_flux, mask=atm.mask)

    # Precision as given by the third condition: M = T**2
    result_3 = rv_precision(wav_grid, sampled_flux, mask=atm.transmission ** 2)

    # Turn quality back into a Quantity (to give it a .value method)
    q = q * u.dimensionless_unscaled
    return q, result_1, result_2, result_3


def convolve_and_resample(
    wav: ndarray,
    flux: ndarray,
    vsini: float,
    R: float,
    band: str,
    sampling: float,
    **conv_kwargs,
) -> Tuple[ndarray, ndarray]:
    """Convolve and resample functions together.

    Returns
    -------
    wav_grid: ndarray
        Resampled wavelength array
    sampled_flux: ndarray
        Convolved and resampled flux array

    """
    wav_band, __, convolved_flux = convolution(wav, flux, vsini, R, band, **conv_kwargs)
    # Re-sample to sampling per resolution element.
    wav_grid = log_resample(wav_band, sampling, R)
    sampled_flux = np.interp(wav_grid, wav_band, convolved_flux)
    return wav_grid, sampled_flux


def model_format_args(model, pars):
    """Format the model and parameter args to save in output file.
    Change to int, str and float.

    model in [temp, logg, fe/h, alpha]
    pars in order (R, band, vsini, sample).

    Can now also optionally handle a 5th parameter RV.

    """
    temp = int(model[0])
    logg = float(model[1])
    fe_h = float(model[2])
    alpha = float(model[3])
    band = pars[1]
    res = int(pars[0] / 1000)
    vsini = float(pars[2])
    sample = float(pars[3])
    try:
        rv = float(pars[4])
        return temp, logg, fe_h, alpha, band, res, vsini, sample, rv
    except IndexError:
        return temp, logg, fe_h, alpha, band, res, vsini, sample, 0.0


def header_row(add_rv=False) -> str:
    """Header row for output file."""
    if add_rv:
        header = (
            "temp,logg,feh,alpha,band,resolution,vsini,sampling,"
            "rv,correctflag,quality,cond1,cond2,cond3\n"
        )
    else:
        header = (
            "temp,logg,feh,alpha,band,resolution,vsini,sampling,"
            "correctflag,quality,cond1,cond2,cond3\n"
        )
    return header


def get_already_computed(filename: str, add_rv: bool = False) -> List[str]:
    """Get the string of already computed model/parameters from the result file."""
    param_lines = []
    with open(filename, "r") as f:
        for line in f:
            if add_rv:
                # First 10 columns
                param_lines.append(select_csv_columns(line, ncols=10))
            else:
                # First 9 columns
                param_lines.append(select_csv_columns(line, ncols=9))
        # Strip any spaces
        param_lines = [strip_whitespace(value) for value in param_lines]
    return param_lines


def is_already_computed(
    computed_values: List[str],
    model,
    pars,
    add_rv: bool = False,
    correct: bool = False,
    verbose=False,
) -> bool:
    """Check if any combinations have already been preformed.
    Correct is boolean for applied Artigau correction."""
    model_par_str_args = model_format_args(model, pars)
    rv_template = "{0:5d},{1:3.01f},{2:4.01f},{3:3.01f},{4:s},{5:3d}k,{6:4.01f},{7:3.01f},{8:3.01f},{9:1d}"
    no_rv_template = (
        "{0:5d},{1:3.01f},{2:4.01f},{3:3.01f},{4:s},{5:3d}k,{6:4.01f},{7:3.01f},{8:1d}"
    )
    if add_rv:
        if len(model_par_str_args) != 9:
            raise ValueError("model_par_str_args is incorrect length")

        idenifying_line = strip_whitespace(
            rv_template.format(*model_par_str_args, int(correct))
        )
    else:
        model_par_str_args = model_par_str_args[:8]
        if len(model_par_str_args) != 8:
            raise ValueError("model_par_str_args is incorrect length")

        idenifying_line = strip_whitespace(
            no_rv_template.format(*model_par_str_args, int(correct))
        )

    result = idenifying_line in computed_values

    if result and verbose:
        print(model_par_str_args, "already computed")
    return result


def strip_whitespace(line: str) -> str:
    return "".join(line.split())


def select_csv_columns(line: str, ncols: int = 8) -> str:
    """Select first ncols in a line from a csv.

    Parameters
    ----------
    ncols: int
        Number of column to select.

    Returns
    -------
    selected_cols: str
        Selected ncols of csv.

    """
    return ",".join(line.split(",")[:ncols])


def check_model(model: str) -> str:
    """Check model is 'aces' or 'btsettl'.

    Parameters
    ----------
    model: str
        Model name. Should be either 'aces' or 'btsettl'.

    Returns
    -------
    model: str
       Valid model output; either 'aces' or 'btsettl'.

"""
    if model == "phoenix":
        warnings.warn(
            "The model name 'phoenix' is depreciated, use 'aces' instead.",
            DeprecationWarning,
        )
        model = "aces"
    if model not in ["aces", "btsettl"]:
        raise ValueError(
            "Model name error of '{}'. Valid choices are 'aces' and 'btsettl'".format(
                model
            )
        )
    return model


if __name__ == "__main__":
    args = _parser()

    # check bt-settl parameters
    if args.model == "btsettl":
        if (args.metal != [0]) or (args.alpha != [0]):
            raise ValueError(
                "You cannot vary metallicity and alpha for BT-Settl, remove these flags."
            )
    try:
        normalize = not args.disable_normalization
    except AttributeError:
        normalize = True

    try:
        num_procs = args.num_procs
    except AttributeError:
        num_procs = num_cpu_minus_1

    conv_kwargs = {
        "epsilon": 0.6,
        "fwhm_lim": 5.0,
        "num_procs": num_procs,
        "normalize": normalize,
        "verbose": args.verbose,
    }

    snr = args.snr
    air = args.air
    if "ALL" in args.bands:
        args.bands.extend(config.bands["all"])
        args.bands = set(args.bands)  # Unique
    ref_band = args.ref_band

    # Load the relevant spectra
    if args.model == "aces":
        models_list = itertools.product(args.temp, args.logg, args.metal, args.alpha)
    else:
        models_list = itertools.product(args.temp, args.logg, [0], [0])

    if (args.rv != [0.0]) and (not args.add_rv):
        raise ValueError("Need to include --add_rv flag if a non-zero RV is used.")

    if not os.path.exists(args.output):
        with open(args.output, "a") as f:
            header = header_row(add_rv=args.add_rv)
            f.write(header)

    # Find all model/parameter combinations already computed.
    # To later skip recalculation.
    computed_values = get_already_computed(args.output, args.add_rv)

    # Iteration over stellar model (temp. logg, feh, alpha) then an inner loop over parameters (R, vsini, sampling, rv)
    for model in models_list:
        # Create generator for params_list
        params_list = itertools.product(
            args.resolution, args.bands, args.vsini, args.sampling, args.rv
        )
        for (R, band, vsini, sample, rv) in params_list:
            pars = (R, band, vsini, sample, rv)

            if is_already_computed(
                computed_values,
                model,
                pars,
                add_rv=args.add_rv,
                correct=args.correct,
                verbose=args.verbose,
            ):
                # skipping the recalculation
                continue
            else:
                precision_result: Tuple[Any, ...] = do_analysis(
                    model,
                    vsini=vsini,
                    R=R,
                    band=band,
                    conv_kwargs=conv_kwargs,
                    snr=snr,
                    ref_band=ref_band,
                    sampling=sample,
                    air=air,
                    model=args.model,
                )
                result: List[Optional[float]] = [
                    round(res.value, 1) if res is not None else None
                    for res in precision_result
                ]

                if args.correct:
                    # Apply Artigau 2018 Corrections
                    corr_value = correct_artigau_2018(band)
                    for ii, res in enumerate(result):
                        if (ii > 0) and (result[ii] is not None):  # Not the quality
                            result[ii] = res * corr_value

                result[0] = int(result[0]) if result[0] is not None else None

                if args.add_rv:
                    output_template = (
                        "{0:5d},{1:3.01f},{2:4.01f},{3:3.01f},{4:s},{5:3d}k,"
                        "{6:4.01f},{7:3.01f},{8:3.01f},{9:1d},{10:6d},{11:5.01f},"
                        "{12:5.01f},{13:5.01f}\n"
                    )
                    output_model_args = model_format_args(model, pars)
                else:
                    output_template = (
                        "{0:5d},{1:3.01f},{2:4.01f},{3:3.01f},{4:s},{5:3d}k,"
                        "{6:4.01f},{7:3.01f},{8:1d},{9:6d},{10:5.01f},{11:5.01f},"
                        "{12:5.01f}\n"
                    )
                    output_model_args = model_format_args(model, pars)[:8]

                linetowite = output_template.format(
                    *output_model_args,
                    int(args.correct),
                    result[0],
                    result[1],
                    result[2],
                    result[3],
                )
                with open(args.output, "a") as f:
                    f.write(strip_whitespace(linetowite) + "\n")  # Make csv only

    if args.verbose:
        print(
            "`{1}` completed successfully:\n"
            "\tYou shall find you results in '{0}'.".format(
                args.output, os.path.basename(__file__)
            )
        )



import numpy as np
import pytest

from eniric.resample import log_resample, wl_logspace


@pytest.mark.parametrize("wav_start, wav_stop", [(2.1, 2.2), (1657.11, 1695)])
@pytest.mark.parametrize("sampling", [2, 5])
@pytest.mark.parametrize("resolution", [20000, 100_001])
def test_log_resample(wav_start, wav_stop, sampling, resolution):
    wav = np.linspace(wav_start, wav_stop, 50)

    resampled = log_resample(wav, sampling, resolution)

    assert resampled[0] == wav[0]
    assert resampled[-1] >= wav[-1]
    assert np.allclose(
        resampled[1:] / resampled[:-1],
        np.ones_like(resampled[:-1]) * (1 + 1 / (sampling * resolution)),
    )


@pytest.fixture(params=[(1, 1.1, 1.01), (1, 10, 1.25)])
def logspace_params(request):
    return request.param


def test_wl_logspace_endpoint(logspace_params):
    start, stop, base = logspace_params
    resample1 = wl_logspace(start, stop, base=base, end_point=False)
    resample2 = wl_logspace(start, stop, base=base, end_point=True)

    assert len(resample1) < len(resample2)

    assert resample1[-1] < stop
    assert resample2[-1] >= stop
    assert resample2[-1] < stop * base ** 2  # Not big


def test_wl_logspace_increases_by_bnd(logspace_params):
    start, stop, base = logspace_params
    resample1 = wl_logspace(start, stop, base=base, end_point=False)
    assert np.allclose(
        resample1[1:] / resample1[:-1], base * np.ones_like(resample1[:-1])
    )



import codecs
import os
import sys
from os.path import join
from typing import List

from setuptools import find_packages, setup

if sys.version < "3.6":
    sys.exit(
        "Error: Python 3.6 or greater required for Eniric (using {})".format(
            sys.version
        )
    )


here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with codecs.open(join(here, "README.md")) as f:
    long_description = f.read()

with codecs.open(join(here, "requirements.txt")) as f:
    requirements = f.read().splitlines()

# Conditional pytest-runner install
needs_pytest = {"pytest", "test", "ptr"}.intersection(sys.argv)
pytest_runner: List[str] = ["pytest-runner>=4"] if needs_pytest else []
setup_requires: List[str] = []

config = {
    "name": "eniric",
    "description": "Eniric: Extended NIR Information Content",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "author": "Jason Neal",
    "url": "https://github.com/jason-neal/eniric.git",
    "download_url": "https://github.com/jason-neal/eniric.git",
    "author_email": "jason.neal@astro.up.pt",
    "version": "1.0rc3",
    "license": "MIT Licence",
    "setup_requires": setup_requires + pytest_runner,
    "tests_require": ["pytest", "hypothesis"],
    "install_requires": requirements,
    "extras_require": {
        "dev": ["check-manifest"],
        "test": ["coverage", "pytest", "pytest-cov", "python-coveralls", "hypothesis"],
        "docs": ["sphinx >= 1.4", "sphinx_rtd_theme", "rstcheck"],
        "ci": [
            "codacy-coverage==1.3.11",
            "codeclimate-test-reporter>=0.2.3",
            "python-coveralls>=2.9.1",
        ],
    },  # $ pip install -e .[dev,test, docs]
    "packages": find_packages(),
    "scripts": [
        "scripts/phoenix_precision.py",
        "scripts/split_atmmodel.py",
        "scripts/barycenter_broaden_atmmodel.py",
        "scripts/untar_here.py",
        "scripts/download/download_eniric_data.sh",
        "scripts/download/ps_download_eniric_data.ps1",
        "scripts/download/download_test_aces.py",
    ],
    "include_package_data": True,
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    "classifiers": [
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Natural Language :: English",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
    ],
    # What does your project relate to?
    "keywords": [
        "Astronomy",
        "Radial velocity",
        "Near-infrared spectroscopy",
        "M-dwarfs",
    ],
}

setup(**config)



#!/usr/bin/env python
"""
csv2tsv.py
----------
Simple script using pandas to transform from \*.csv to \*.tsv.

"""

import argparse
import sys

import pandas


def _parser():
    """Take care of all the argparse stuff."""
    parser = argparse.ArgumentParser(description="Transform a *.csv file into a *.tsv.")

    parser.add_argument("name", help="Filename to transform.", type=str, default="")
    return parser.parse_args()


if __name__ == "__main__":
    args = _parser()
    name = args.name

    df = pandas.read_csv(name, header=0)
    df.to_csv(name.replace(".csv", ".tsv"), sep="\t", header=True, index=False)
    sys.exit(0)



"""Tests for functions and bits of code from the scripts in 'scripts'."""

import argparse

import pytest

from scripts.split_atmmodel import check_positive


def test_check_positive():
    """Test that positive string values are returned as floats and negative values as errors."""
    assert check_positive("1") == 1.0
    assert isinstance(check_positive("20"), float)


@pytest.mark.parametrize(
    "val,error",
    [("-1", argparse.ArgumentTypeError), (-1.0, ValueError), (10, ValueError)],
)  # Not a float
def test_check_positive_errors(val, error):
    with pytest.raises(error):
        check_positive(val)



"""
precision_four_panel.py
-----------------------
Plot a Figueira et al. (2016) Figure 1 like plot.

"""
import argparse
import sys
from os.path import join

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from eniric import config
from eniric.utilities import rv_cumulative, rv_cumulative_full

matplotlib.use("Agg")


def load_dataframe(filename):
    """Load in precision file, clean up spaces in csv.

    Parameters
    ----------
    filename: str
        Name of phoenix_precision.py output.

    Returns
    -------
    df: pandas.DataFrame
        DataFrame of data.

    """
    df = pd.read_csv(precision_file)
    # Temp, logg, [Fe/H], Alpha, Band, Resolution, vsini, Sampling, Quality, Cond. 1, Cond. 2, Cond. 3, correct flag
    df.columns = df.columns.str.strip()
    df.Band = df.Band.str.strip()
    df.Resolution = df.Resolution.str.strip()
    df.Quality = df.Quality.astype(float)
    df.Temp = df.Temp.astype(float)
    df.logg = df.logg.astype(float)
    df["[Fe/H]"] = df["[Fe/H]"].astype(float)
    df.Alhpa = df.Alpha.astype(float)
    df.vsini = df.vsini.astype(float)
    df.Sampling = df.Sampling.astype(float)
    df["Cond. 1"] = df["Cond. 1"].astype(float)
    df["Cond. 2"] = df["Cond. 2"].astype(float)
    df["Cond. 3"] = df["Cond. 3"].astype(float)
    df["correct flag"] = df["correct flag"].astype(bool)

    return df


def plot_precision(
    precision_file, teffs=None, logg=4.5, fe_h=0.0, vsini=1.0, sampling=3
):
    """Plot precision 4 panel with RV  precision.

    Saves figure to ``plots/``.

    Parameters
    ----------
    precision_file: str
        Name of phoenix_precision.py output.
    teffs: List of int or None
       Stellar temperatures. Default is [3900, 3500, 2800, 2600].
    logg: int
        Stellar Logg. Default is 4.5.
    fe_h: int
        Stellar metallicity. Default is 0.0.
    vsini: float
       Rotational velocity. Default is 1.0.
    sampling:
        Spectral sampling. Default is 3.

    """
    if teffs is None:
        # Default teffs
        teffs = [3900, 3500, 2800, 2600]
    assert len(teffs) == 4
    df = load_dataframe(precision_file)
    filter_dict = {"logg": logg, "[Fe/H]": fe_h, "vsini": vsini, "Sampling": sampling}
    df = filter_df(
        df, filter_dict, drop_list=["Alpha", "[Fe/H]", "correct flag", "Quality"]
    )

    fig, axes = plt.subplots(2, 2)
    ax = axes.flatten()
    df_selected = df[df.Resolution.isin(["60k", "80k", "100k"])]
    df_selected = df_selected[df_selected.Temp.isin(teffs)]

    maximums = []
    minimums = []

    for ii, temp in enumerate(teffs):
        # This entry
        df_ii = df_selected[df_selected["Temp"] == temp]
        df_ii_60k = df_ii[df_ii["Resolution"].str.strip() == "60k"]
        df_ii_80k = df_ii[df_ii["Resolution"].str.strip() == "80k"]
        df_ii_100k = df_ii[df_ii["Resolution"].str.strip() == "100k"]

        df_ii_60k = df_ii_60k.set_index("Band")
        df_ii_60k = df_ii_60k.reindex(["Z", "Y", "J", "H", "K"])

        df_ii_80k = df_ii_80k.set_index("Band")
        df_ii_80k = df_ii_80k.reindex(["Z", "Y", "J", "H", "K"])
        df_ii_100k = df_ii_100k.set_index("Band")
        df_ii_100k = df_ii_100k.reindex(["Z", "Y", "J", "H", "K"])

        maximums.append(
            np.max(
                [
                    df_ii_60k[["Cond. 1", "Cond. 2", "Cond. 3"]].max(),
                    df_ii_80k[["Cond. 1", "Cond. 2", "Cond. 3"]].max(),
                    df_ii_100k[["Cond. 1", "Cond. 2", "Cond. 3"]].max(),
                ]
            )
        )
        minimums.append(
            np.min(
                [
                    df_ii_60k[["Cond. 1", "Cond. 2", "Cond. 3"]].min(),
                    df_ii_80k[["Cond. 1", "Cond. 2", "Cond. 3"]].min(),
                    df_ii_100k[["Cond. 1", "Cond. 2", "Cond. 3"]].min(),
                ]
            )
        )

        ax[ii].fill_between(
            df_ii_60k.index,
            df_ii_60k["Cond. 2"].values,
            df_ii_60k["Cond. 3"].values,
            color="b",
            alpha=0.2,
        )
        ax[ii].fill_between(
            df_ii_80k.index,
            df_ii_80k["Cond. 2"].values,
            df_ii_80k["Cond. 3"].values,
            color="g",
            alpha=0.2,
        )
        ax[ii].fill_between(
            df_ii_100k.index,
            df_ii_100k["Cond. 2"].values,
            df_ii_100k["Cond. 3"].values,
            color="r",
            alpha=0.2,
        )

        ax[ii].plot(
            df_ii_60k.index, df_ii_60k["Cond. 1"].values, color="b", linestyle="--"
        )  # lim
        ax[ii].plot(
            df_ii_80k.index, df_ii_80k["Cond. 1"].values, color="g", linestyle="--"
        )  # lim
        ax[ii].plot(
            df_ii_100k.index, df_ii_100k["Cond. 1"].values, color="r", linestyle="--"
        )  # lim

        ax[ii].scatter(
            df_ii_60k.index,
            df_ii_60k["Cond. 2"].values,
            marker="^",
            color="b",
            alpha=0.4,
        )
        ax[ii].scatter(
            df_ii_60k.index,
            df_ii_60k["Cond. 3"].values,
            marker="o",
            color="b",
            alpha=0.4,
        )

        ax[ii].scatter(
            df_ii_80k.index,
            df_ii_80k["Cond. 3"].values,
            marker="^",
            color="g",
            alpha=0.4,
        )
        ax[ii].scatter(
            df_ii_80k.index,
            df_ii_80k["Cond. 2"].values,
            marker="o",
            color="g",
            alpha=0.4,
        )

        ax[ii].scatter(
            df_ii_100k.index,
            df_ii_100k["Cond. 3"].values,
            marker="^",
            color="r",
            alpha=0.4,
        )
        ax[ii].scatter(
            df_ii_100k.index,
            df_ii_100k["Cond. 2"].values,
            marker="o",
            color="r",
            alpha=0.4,
        )

    # Set limits ticks and labels
    ymax = np.max(maximums)
    ymin = np.min(minimums)
    delta_y = ymax - ymin
    band_size = len(df_ii_60k.index)
    for jj in range(4):
        ax[jj].text(0, ymax, "{} K".format(teffs[jj]), size=14)
        ax[jj].set_ylim(ymin - 0.11 * delta_y, ymax + 0.11 * delta_y)
        ax[jj].set_xlim(-0.5, band_size - 0.5)
        ax[jj].tick_params(axis="both", which="major", labelsize=12)

        # ticks and labels
        if (jj == 2) or (jj == 3):
            ax[jj].set_xlabel("Bands", fontsize=12)
        if (jj == 1) or (jj == 3):
            ax[jj].set_yticklabels([])
        if (jj == 0) or (jj == 1):
            ax[jj].set_xticklabels([])

    fig.text(
        0.04,
        0.5,
        r"Precision [m/s]",
        ha="center",
        va="center",
        rotation="vertical",
        size=12,
    )

    fig.subplots_adjust(hspace=0, wspace=0, bottom=0.12, top=0.95, right=0.95)

    fig.savefig("plots/precision_logg{0}_feh_{1}.pdf".format(logg, fe_h))
    fig.savefig("plots/precision_logg{0}_feh_{1}.png".format(logg, fe_h), dpi=400)


def filter_df(df, filter_dict, drop_list=None):
    """Filter DataFrame by dictionary of key and values."""
    for key, val in filter_dict.items():
        df = df[df[key] == val]
    if drop_list is not None:
        df = df.drop(columns=drop_list)
    return df


def cumulative_df(df, full_cum=False):
    """Calculated cumulative RV precision across bands.
    The precision of "Z", "ZY", "ZYJ", "ZYJH", "ZYJHK" bands.

    Parameters
    ----------
    df: pandas.DataFrame
       DataFrame.
    full_cum: bool
       Include "YJHK", "JHK", "HK", "K" grouping also. Default is False.

    """
    bands = df.index
    assert all(bands == ["Z", "Y", "J", "H", "K"]), bands
    if full_cum:
        cum_bands = ["Z", "ZY", "ZYJ", "ZYJH", "ZYJHK", "YJHK", "JHK", "HK", "K"]

        cum_dict = {
            "Band": cum_bands,
            "Cond. 1": rv_cumulative_full(df["Cond. 1"]),
            "Cond. 2": rv_cumulative_full(df["Cond. 2"]),
            "Cond. 3": rv_cumulative_full(df["Cond. 3"]),
        }
    else:
        cum_bands = ["Z", "ZY", "ZYJ", "ZYJH", "ZYJHK"]
        cum_dict = {
            "Band": cum_bands,
            "Cond. 1": rv_cumulative(df["Cond. 1"], single=True),
            "Cond. 2": rv_cumulative(df["Cond. 2"], single=True),
            "Cond. 3": rv_cumulative(df["Cond. 3"], single=True),
        }
    cum_df = pd.DataFrame(cum_dict)
    cum_df = cum_df.set_index("Band")
    cum_df = cum_df.reindex(cum_bands)
    return cum_df


def cumulative_plot(
    precision_file,
    teffs=None,
    logg=4.5,
    fe_h=0.0,
    vsini=1.0,
    sampling=3,
    full_cum=False,
):
    """RV precision with cumulative bands.

    full_cum: bool
    Cumlative over entire range [ "Z","ZY", "ZYJ", "ZYJH", "ZYJHK","YJHK", "JHK","HK","K"]

    Saves figure to ``plots/``.

    Parameters
    ----------
    precision_file: str
        Name of phoenix_precision.py output.
    teffs: List of int or None
       Stellar temperatures. Default is [3900, 3500, 2800, 2600].
    logg: int
        Stellar Logg. Default is 4.5.
    fe_h: int
        Stellar metallicity. Default is 0.0.
    vsini: float
       Rotational velocity. Default is 1.0.
    sampling:
        Spectral sampling. Default is 3.
    full_cum: bool
        Cumulative over entire range. Default is False.

    """
    if teffs is None:
        # Default values
        teffs = [3900, 3500, 2800, 2600]
    assert len(teffs) == 4

    df = load_dataframe(precision_file)
    filter_dict = {"logg": logg, "[Fe/H]": fe_h, "vsini": vsini, "Sampling": sampling}
    df = filter_df(
        df, filter_dict, drop_list=["Alpha", "[Fe/H]", "correct flag", "Quality"]
    )

    fig, axes = plt.subplots(2, 2)
    ax = axes.flatten()
    df_selected = df[df.Resolution.isin(["60k", "80k", "100k"])]
    df_selected = df_selected[df_selected.Temp.isin(teffs)]

    maximums = []
    minimums = []

    for ii, temp in enumerate(teffs):
        # This entry
        df_ii = df_selected[df_selected["Temp"] == temp]
        df_ii_60k = df_ii[df_ii["Resolution"].str.strip() == "60k"]
        df_ii_80k = df_ii[df_ii["Resolution"].str.strip() == "80k"]
        df_ii_100k = df_ii[df_ii["Resolution"].str.strip() == "100k"]

        df_ii_60k = df_ii_60k.set_index("Band")
        df_ii_60k = df_ii_60k.reindex(["Z", "Y", "J", "H", "K"])

        df_ii_80k = df_ii_80k.set_index("Band")
        df_ii_80k = df_ii_80k.reindex(["Z", "Y", "J", "H", "K"])
        df_ii_100k = df_ii_100k.set_index("Band")
        df_ii_100k = df_ii_100k.reindex(["Z", "Y", "J", "H", "K"])

        # Cumulative
        df_ii_60k = cumulative_df(df_ii_60k, full_cum=full_cum)
        df_ii_80k = cumulative_df(df_ii_80k, full_cum=full_cum)
        df_ii_100k = cumulative_df(df_ii_100k, full_cum=full_cum)

        maximums.append(
            np.max(
                [
                    df_ii_60k[["Cond. 1", "Cond. 2", "Cond. 3"]].max(),
                    df_ii_80k[["Cond. 1", "Cond. 2", "Cond. 3"]].max(),
                    df_ii_100k[["Cond. 1", "Cond. 2", "Cond. 3"]].max(),
                ]
            )
        )
        minimums.append(
            np.min(
                [
                    df_ii_60k[["Cond. 1", "Cond. 2", "Cond. 3"]].min(),
                    df_ii_80k[["Cond. 1", "Cond. 2", "Cond. 3"]].min(),
                    df_ii_100k[["Cond. 1", "Cond. 2", "Cond. 3"]].min(),
                ]
            )
        )

        ax[ii].fill_between(
            df_ii_60k.index,
            df_ii_60k["Cond. 2"].values,
            df_ii_60k["Cond. 3"].values,
            color="b",
            alpha=0.2,
        )
        ax[ii].fill_between(
            df_ii_80k.index,
            df_ii_80k["Cond. 2"].values,
            df_ii_80k["Cond. 3"].values,
            color="g",
            alpha=0.2,
        )
        ax[ii].fill_between(
            df_ii_100k.index,
            df_ii_100k["Cond. 2"].values,
            df_ii_100k["Cond. 3"].values,
            color="r",
            alpha=0.2,
        )

        ax[ii].plot(
            df_ii_60k.index, df_ii_60k["Cond. 1"].values, color="b", linestyle="--"
        )  # lim
        ax[ii].plot(
            df_ii_80k.index, df_ii_80k["Cond. 1"].values, color="g", linestyle="--"
        )  # lim
        ax[ii].plot(
            df_ii_100k.index, df_ii_100k["Cond. 1"].values, color="r", linestyle="--"
        )  # lim

        ax[ii].scatter(
            df_ii_60k.index,
            df_ii_60k["Cond. 2"].values,
            marker="^",
            color="b",
            alpha=0.4,
        )
        ax[ii].scatter(
            df_ii_60k.index,
            df_ii_60k["Cond. 3"].values,
            marker="o",
            color="b",
            alpha=0.4,
        )

        ax[ii].scatter(
            df_ii_80k.index,
            df_ii_80k["Cond. 3"].values,
            marker="^",
            color="g",
            alpha=0.4,
        )
        ax[ii].scatter(
            df_ii_80k.index,
            df_ii_80k["Cond. 2"].values,
            marker="o",
            color="g",
            alpha=0.4,
        )

        ax[ii].scatter(
            df_ii_100k.index,
            df_ii_100k["Cond. 3"].values,
            marker="^",
            color="r",
            alpha=0.4,
        )
        ax[ii].scatter(
            df_ii_100k.index,
            df_ii_100k["Cond. 2"].values,
            marker="o",
            color="r",
            alpha=0.4,
        )

        ax[ii].fill_between(
            df_ii_100k.index,
            df_ii_100k["Cond. 2"].values,
            df_ii_100k["Cond. 3"].values,
            color="r",
            alpha=0.2,
        )

    # Set limits ticks and labels
    ymax = np.min([np.max(maximums), 10])  # Set limit to 20 km/s
    ymin = np.min(minimums)
    delta_y = ymax - ymin
    band_size = len(df_ii_60k.index)

    for jj in range(4):
        ax[jj].text(0, ymax, "{} K".format(teffs[jj]), size=14)
        ax[jj].set_ylim(ymin - 0.1 * delta_y, ymax + 0.15 * delta_y)
        ax[jj].set_xlim(-0.5, band_size - 0.5)
        ax[jj].tick_params(axis="both", which="major", labelsize=12)

        # ticks and labels
        if (jj == 2) or (jj == 3):
            ax[jj].set_xlabel("Bands", fontsize=12)
            if full_cum:
                ax[jj].tick_params(axis="x", labelrotation=40)
            else:
                ax[jj].tick_params(axis="x", labelrotation=25)

        if (jj == 1) or (jj == 3):
            ax[jj].set_yticklabels([])
        if (jj == 0) or (jj == 1):
            ax[jj].set_xticklabels([])

    fig.text(
        0.04,
        0.5,
        r"Precision [m/s]",
        ha="center",
        va="center",
        rotation="vertical",
        size=12,
    )

    fig.subplots_adjust(hspace=0, wspace=0, bottom=0.17, top=0.95, right=0.95)

    fname = "plots/cummulative_precision_logg{0}_feh_{1}_{2}".format(
        logg, fe_h, full_cum
    )
    fig.savefig(fname + ".pdf")
    fig.savefig(fname + ".png", dpi=400)


default_file = join(
    config.pathdir, config.paths["precision_results"], "precision_results.csv"
)


def _parser():
    """Take care of all the argparse stuff."""
    parser = argparse.ArgumentParser(
        description="Create a four panel RV relative precision plot."
    )
    parser.add_argument(
        "precision_file",
        help="Precision result csv to use. Default is set with the config.yaml file. Currently {0}".format(
            default_file
        ),
        type=str,
        default=default_file,
    )
    parser.add_argument(
        "-t",
        "--temperatures",
        nargs=4,
        help="Temperatures to display in Kelvin. Default is [3900, 3500, 2800, 2600].",
        type=int,
        default=[3900, 3500, 2800, 2600],
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = _parser()
    precision_file = args.precision_file
    temperatures = args.temperatures

    plot_precision(precision_file, teffs=temperatures)
    # plot_precision(precision_file, teffs=temperatures, logg=4, fe_h=1)
    cumulative_plot(precision_file, teffs=temperatures)
    cumulative_plot(precision_file, teffs=temperatures, full_cum=True)
    sys.exit(0)



import pytest

from scripts.phoenix_precision import select_csv_columns, strip_whitespace, check_model


@pytest.mark.parametrize(
    "initial, expected",
    [
        (" test string goes here ", "teststringgoeshere"),
        (" 5500,  4.50, 0.0, K, 43.2322 , ", "5500,4.50,0.0,K,43.2322,"),
    ],
)
def test_strp_space(initial, expected):
    assert expected == strip_whitespace(initial)


def test_select_csv_columns():
    assert select_csv_columns("1,2,3,4,5,6,7,8,9,10") == "1,2,3,4,5,6,7,8"


@pytest.mark.parametrize(
    "initial, ncols, expected",
    [
        ("1,2,3,4,5,6,7", 4, "1,2,3,4"),
        ("hello,2,3,4,5,6,7,8,there, 10", 9, "hello,2,3,4,5,6,7,8,there"),
    ],
)
def test_select_csv_columns_with_ncols(initial, ncols, expected):
    assert select_csv_columns(initial, ncols=ncols) == expected


@pytest.mark.parametrize(
    "model_in, expected_model",
    [
        ("aces", "aces"),
        ("phoenix", "aces"),
        ("btsettl", "btsettl"),
    ],
)
def test_check_model(model_in, expected_model):
    model_out = check_model(model_in)
    assert model_out == expected_model


def test_model_depreciation():
    with pytest.deprecated_call():
        check_model("phoenix")


@pytest.mark.parametrize(
    "invalid_model",
    ["phoneix-aces", "best", 42],
)
def test_check_model_error(invalid_model):
    with pytest.raises(ValueError):
        check_model(invalid_model)
